"""
============
Align Single
============

Plot a single frequency channel of raw rf data and the result of
interpolation and Savitzky–Golay smoothing. Save plot to ``./embers_out/condition_data``

"""

import argparse
import pkg_resources
from pathlib import Path
import matplotlib.pyplot as plt
from embers.condition_data.align_data import savgol_interp

_parser = argparse.ArgumentParser(
    description="""
    Create sample savgol interp plot of pair of rf data
    """
)

_parser.add_argument(
    "--ref_file", metavar="\b", default="", help="Reference rf data file"
)

_parser.add_argument("--tile_file", metavar="\b", default="", help="Tile rf data file")

_parser.add_argument(
    "--savgol_window_1",
    metavar="\b",
    type=int,
    default=11,
    help="First savgol window. Default=11",
)

_parser.add_argument(
    "--savgol_window_2",
    metavar="\b",
    type=int,
    default=15,
    help="Second savgol window. Default=15",
)

_parser.add_argument(
    "--polyorder", metavar="\b", type=int, default=2, help="Polynomial order. Default=2"
)

_parser.add_argument(
    "--interp_type",
    metavar="\b",
    default="cubic",
    help="Interpolation type. Default=cubic",
)

_parser.add_argument(
    "--interp_freq",
    metavar="\b",
    type=int,
    default=1,
    help="Interpolation frequency. Default=1",
)

_parser.add_argument(
    "--channel", metavar="\b", default="", help="Frequency channel to plot"
)

_parser.add_argument(
    "--out_dir",
    metavar="\b",
    default="./embers_out/condition_data",
    help="Dir where savgol-interp sample plot is saved. Default=./embers_out/condition_data",
)

_args = _parser.parse_args()
_ref_file = _args.ref_file
_tile_file = _args.tile_file
_savgol_window_1 = _args.savgol_window_1
_savgol_window_2 = _args.savgol_window_2
_polyorder = _args.polyorder
_interp_type = _args.interp_type
_interp_freq = _args.interp_freq
_channel = _args.channel
_out_dir = _args.out_dir


# if no input files provided, use sample package data
if _ref_file == "":
    print("----------------------------------------------------------")
    print("No ref_file provided, using packaged sample data")
    _ref_file = pkg_resources.resource_filename(
        "embers.kindle", "data/rf_data/rf0XX/2019-10-10/rf0XX_2019-10-10-02:30.txt"
    )

if _tile_file == "":
    print("No tile_file provided, using packaged sample data")
    _tile_file = pkg_resources.resource_filename(
        "embers.kindle", "data/rf_data/S06XX/2019-10-10/S06XX_2019-10-10-02:30.txt"
    )

if _channel == "":
    print("No frequency channel provided, using 59 for sample data")
    print("")
    print(">>> savgol_interp_sample --help, for more options")
    print("----------------------------------------------------------")
    _channel = 59


def plot_savgol_interp(
    ref=None,
    tile=None,
    savgol_window_1=None,
    savgol_window_2=None,
    polyorder=None,
    interp_type=None,
    interp_freq=None,
    channel=None,
    out_dir=None,
):

    (
        ref_ali,
        tile_ali,
        time_array,
        ref_power,
        tile_power,
        ref_time,
        tile_time,
    ) = savgol_interp(
        ref=ref,
        tile=tile,
        savgol_window_1=savgol_window_1,
        savgol_window_2=savgol_window_2,
        polyorder=polyorder,
        interp_type=interp_type,
        interp_freq=interp_freq,
    )

    # Sample align plot
    plt.style.use("seaborn")
    plt.rcParams["figure.figsize"] = (9, 6)

    # convert times to minuts from first datapoint
    time_array = (time_array - time_array[0]) / 60
    ref_time = (ref_time - ref_time[0]) / 60
    tile_time = (tile_time - tile_time[0]) / 60

    plt.plot(
        time_array,
        tile_ali[::, channel],
        color="#e23a4e",
        alpha=0.9,
        label="tile savgol",
    )
    plt.scatter(
        tile_time,
        tile_power[::, channel],
        color="#f78b51",
        marker=".",
        alpha=0.6,
        label="tile raw",
    )

    plt.plot(
        time_array,
        ref_ali[::, channel],
        color="#252b40",
        alpha=0.9,
        label="ref savgol",
    )
    plt.scatter(
        ref_time,
        ref_power[::, channel],
        color="#6a82bb",
        marker=".",
        alpha=0.6,
        label="ref raw",
    )

    leg = plt.legend(loc="upper left", frameon=True)
    leg.get_frame().set_facecolor("white")
    for l in leg.legendHandles:
        l.set_alpha(1)

    plt.ylim(-110, -20)
    plt.ylabel("Raw Power [dBm]")
    plt.xlabel("Time [min]")
    plt.tight_layout()
    Path(f"{out_dir}").mkdir(parents=True, exist_ok=True)
    plt.savefig(f"{out_dir}/savgol_interp_sample.png")


def main():
    """Execute align_single from terminal."""

    print(f"Saving sample savgol_interp plot to: {_out_dir}")
    plot_savgol_interp(
        ref=_ref_file,
        tile=_tile_file,
        savgol_window_1=_savgol_window_1,
        savgol_window_2=_savgol_window_2,
        polyorder=_polyorder,
        interp_type=_interp_type,
        interp_freq=_interp_freq,
        channel=_channel,
        out_dir=_out_dir,
    )



"""
=======
Colormaps
=======

Visualise custom colormaps used by embers.
Creates sample plot of ember colormaps 
saved to ``./embers_out/condition_data/colormaps.png``

"""

import argparse
import numpy as np
from pathlib import Path
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from embers.condition_data.colormaps import spectral, jade

_spec, _spec_r = spectral()
_jade, _jade_r = jade()


parser = argparse.ArgumentParser(
    description="""
    Visualization of custom ember colormaps
    """
)

parser.add_argument(
    "--out_dir",
    metavar="\b",
    default="./embers_out/condition_data",
    help="Dir where colormap sample plot is saved. Default=./embers_out/condition_data",
)

args = parser.parse_args()
out_dir = Path(args.out_dir)

# Make outdir if it doesn't exist
out_dir.mkdir(parents=True, exist_ok=True)


def waves_2d():
    """Create 2d sine wave.

    Returns
    -------
    :return:
        - sine2d- 2d sine wave
    :rtype: (float, numpy.array(float))

    """

    xx, yy = np.meshgrid(np.linspace(0, 3 * np.pi, 512), np.linspace(0, 3 * np.pi, 512))
    sine2d = np.sin(xx) + np.sin(yy)
    return sine2d


def plt_colormaps(spec, spec_r, jade, jade_r, out_dir):
    """Plot 2x2 grid of sample colormaps.
    
    Parameters
    ----------
    :param ember_cmaps: custom ember colormaps
    :type ember_cmaps: `~matplotlib.colors.Colormap`
    :param out_dir: path to output directory
    :type out_dir: str

    Returns
    -------
    :return: waterfall plot saved by `matplotlib.pyplot.savefig`

    """

    fig = plt.figure(figsize=(9, 8))

    ax1 = fig.add_subplot(221)
    ax1.set_title("Spectral colormap")
    im1 = ax1.imshow(waves_2d(), origin="lower", interpolation="none", cmap=_spec)
    divider = make_axes_locatable(ax1)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im1, cax=cax, orientation="vertical")
    ax1.set_xticklabels([])

    ax2 = fig.add_subplot(222)
    ax2.set_title("Jade colormap")
    im2 = ax2.imshow(waves_2d(), origin="lower", interpolation="none", cmap=_jade)
    divider = make_axes_locatable(ax2)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im2, cax=cax, orientation="vertical")
    ax2.set_xticklabels([])
    ax2.set_yticklabels([])

    ax3 = fig.add_subplot(223)
    ax3.set_title("Spectral_r colormap")
    im3 = ax3.imshow(waves_2d(), origin="lower", interpolation="none", cmap=_spec_r)

    divider = make_axes_locatable(ax3)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im3, cax=cax, orientation="vertical")

    ax4 = fig.add_subplot(224)
    ax4.set_title("Jade_r colormap")
    im4 = ax4.imshow(waves_2d(), origin="lower", interpolation="none", cmap=_jade_r)
    divider = make_axes_locatable(ax4)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(im4, cax=cax, orientation="vertical")
    ax4.set_yticklabels([])

    plt.tight_layout()
    plt.savefig(f"{out_dir}/colormaps.png")


def main():
    """Execute colormaps from terminal."""

    print(f"Plot of embers colormaps saved to ./{out_dir}/colormaps.png")
    plt_colormaps(spec, spec_r, jade, jade_r, out_dir)



"""
===============
Waterfall Batch
===============

Create waterfall plots for all rf data files recorded within a date interval.
Output files are saved to ``./embers_out/condition_data/waterfalls``

"""

import logging
import argparse
import pkg_resources
from pathlib import Path
import concurrent.futures
from itertools import repeat
from embers.condition_data.rf_data import tile_names, time_tree, batch_waterfall


_parser = argparse.ArgumentParser(
    description="""
    Create waterfall plots for all rf_files within a date interval
    """
)

_parser.add_argument(
    "--start_date", metavar="\b", default="", help="start date in YYYY-MM-DD format"
)

_parser.add_argument(
    "--stop_date", metavar="\b", default="", help="stop date in YYYY-MM-DD format"
)

_parser.add_argument(
    "--data_dir", metavar="\b", default="", help="root of dir where rf data is saved"
)

_parser.add_argument(
    "--out_dir",
    metavar="\b",
    default="./embers_out/condition_data",
    help="Dir where colormap sample plot is saved. Default=./embers_out/condition_data",
)

_args = _parser.parse_args()
_start_date = _args.start_date
_stop_date = _args.stop_date
_data_dir = _args.data_dir
_out_dir = _args.out_dir

# if no input file provided, use sample package data
if _data_dir == "":
    print("----------------------------------------------------------")
    print("No data dir provided, using packaged sample data")
    _data_dir = pkg_resources.resource_filename("embers.kindle", "data/rf_data/")

if _start_date == "":
    print("No start_date provided, using 2019-10-01 for sample data")
    _start_date = "2019-10-01"

if _stop_date == "":
    print("No stop_date provided, using 2019-10-10 for sample data")
    print("")
    print(">>> waterfall_batch --help, for more options")
    print("----------------------------------------------------------")
    _stop_date = "2019-10-10"


# Logging config
_log_dir = Path(f"{_out_dir}/waterfalls")
_log_dir.mkdir(parents=True, exist_ok=True)
logging.basicConfig(
    filename=f"{_out_dir}/waterfalls/waterfall_batch.log",
    level=logging.INFO,
    format="%(levelname)s: %(funcName)s: %(message)s",
)


def waterfall_batch(start_date, stop_date, data_dir, out_dir):
    """
    Save a series of waterfall plots in parallel.

    Parameters
    ----------
    :param start_date: date in style YYYY-MM-DD
    :type start_date: str
    :param stop_date: date in style YYYY-MM-DD
    :type stop_date: str
    :param data_dir: path to root of rf data dir
    :type data_dir: str
    :param out_dir: path to output dir
    :type out_dir: str
    
    """

    dates, time_stamps = time_tree(start_date, stop_date)

    for tile in tile_names():
        for day in range(len(dates)):

            with concurrent.futures.ProcessPoolExecutor() as executor:
                results = executor.map(
                    batch_waterfall,
                    repeat(tile),
                    time_stamps[day],
                    repeat(data_dir),
                    repeat(out_dir),
                )

            for result in results:
                logging.info(result)


def main():
    """Execute waterfall from terminal."""

    print(f"Processing rf data files between {_start_date} and {_stop_date}")
    print(f"Saving waterfall plots to: ./{_log_dir}")
    waterfall_batch(_start_date, _stop_date, _data_dir, _out_dir)



import sys
import math
import json
import pytz
import numpy as np

from pathlib import Path
from scipy import interpolate
from datetime import datetime, timedelta


def obs_times(time_zone, start_date, stop_date):
    """Time stuff for 30 minute observations

    returns:
        obs_time        : YYYY-MM-DD-HH:MM of start of 30 min observations
        obs_unix        : Start of 30 min obs in unix time
        obs_unix_end    : End of 30 min obs in unix time
    """

    # Time stuff
    # The time input is in local time. As in Austraila/Perth
    local = pytz.timezone(time_zone)

    t_start = datetime.strptime(start_date, "%Y-%m-%d")
    t_stop = datetime.strptime(stop_date, "%Y-%m-%d")

    # Number of days that the date range spans
    n_days = (t_stop - t_start).days

    # YYYY-MM-DD-HH:MM format
    obs_time = []

    # Start of half hour obs in unix time
    obs_unix = []

    # End of half hour obs in unix time
    obs_unix_end = []

    # The +1 makes the date ranges inclusive
    for i in range(n_days + 1):
        day = t_start + timedelta(days=i)
        date = day.strftime("%Y-%m-%d")

        # loop over 48x30 minute obs in a day
        for j in range(48):
            t_delta = datetime.strptime(date, "%Y-%m-%d") + timedelta(minutes=30 * j)
            # convert t_delta to a readable string YYYY-MM-DD-HH:MM
            d_time = t_delta.strftime("%Y-%m-%d-%H:%M")

            # Convert from naive local time to utc aware time
            utc_delta = local.localize(t_delta, is_dst=None).astimezone(pytz.utc)

            # convert to a unix timestamp, used within rf explorer data files
            utc_unix = utc_delta.timestamp()
            # time at end of half hour window
            utc_unix_end = utc_unix + (30 * 60)

            obs_time.append(d_time)
            obs_unix.append(utc_unix)
            obs_unix_end.append(utc_unix_end)

    return (obs_time, obs_unix, obs_unix_end)


def write_json(data, filename=None, out_dir=None):
    """writes data to json file in output dir"""

    with open(f"{out_dir}/{filename}", "w") as f:
        json.dump(data, f, indent=4)


def interp_ephem(t_array, s_alt, s_az, interp_type, interp_freq):
    """Interpolates satellite ephemeris - time, alt, az
    
    Interpolate sat ephem to same freq as align data. This ensures
    that each point in the data, will have an corresponding ephem
    point.

    Args:
        t_array:        Time array of on satellite pass
        s_alt, s_az:    Satellite alt, az at the t_array
        interp_type:    Type of interpolation. Ex: Cubic,Linear. Default=Cubic'
        interp_freq:    Frequency at which to interpolate, in Hertz. 
        
    Returns:
        time_interp:    Interpolated time array
        sat_alt:        Interpolated alt array
        sat_az:         Interpolated az array
    """

    # Create interpolation functions. Math functions, not Python!
    alt_interp = interpolate.interp1d(t_array, s_alt, kind=interp_type)

    # The next step was a bit tricky. Azimuth may wrap around.
    # This may lead to a discontinuity between 0, 2π
    # We deal with this, by unwrapping the angles

    # This extends the angles beyond 2π, if the angles cross the discontinuity
    s_az_cont = np.unwrap(s_az)
    az_interp = interpolate.interp1d(t_array, s_az_cont, kind=interp_type)

    # Makes start and end times clean integers
    # Also ensures that the interp range is inclusive of data points
    start = math.ceil(t_array[0])
    stop = math.floor(t_array[-1])

    # Create time array, at which to evaluate alt/az of sat
    time_interp = list(np.double(np.arange(start, stop, (1 / interp_freq))))

    sat_alt = list(alt_interp(time_interp))

    # The modulus division by 2π, un-does the np.unwrap
    sat_az = list(az_interp(time_interp) % (2 * np.pi))

    return (time_interp, sat_alt, sat_az)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="""
            Collates satellite pass data from all ephem json files.
            The native skyfiled gps timestamps are converted to unix
            timestamps to match the output of the rf explorers. The 
            alt, az data is interpolated to match the cadence of 
            align_data.py. Make a json file with all the passes 
            from each 30 min observation. This will help in the 
            next stage, where we identify all sats in each obs.
            """
    )

    parser.add_argument(
        "--json_dir",
        metavar="\b",
        default="./../../outputs/sat_ephemeris/ephem_json/",
        help="Directory where ephem json files live. Default=./../../outputs/sat_ephemeris/ephem_json/",
    )
    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/sat_ephemeris/chrono_json/",
        help="Output directory. Default=./../../outputs/sat_ephemeris/chrono_json/",
    )
    parser.add_argument(
        "--interp_type",
        metavar="\b",
        default="cubic",
        help="Type of interpolation. Ex: Cubic,Linear. Default=cubic",
    )
    parser.add_argument(
        "--interp_freq",
        metavar="\b",
        type=int,
        default=1,
        help="Frequency at which to interpolate, in Hertz. Must be the same as used in align_data.py. Default=1",
    )
    parser.add_argument(
        "--start_date",
        metavar="\b",
        help="Date from which to determine sat ephemeris. Ex: 2019-10-10",
    )
    parser.add_argument(
        "--stop_date",
        metavar="\b",
        help="Date until which to determine sat ephemeris. Ex: 2019-10-11",
    )
    parser.add_argument(
        "--time_zone",
        metavar="\b",
        default="Australia/Perth",
        help="Time zone where data was recorded. Default=Australia/Perth",
    )

    args = parser.parse_args()

    json_dir = args.json_dir
    out_dir = args.out_dir
    interp_type = args.interp_type
    interp_freq = args.interp_freq
    start_date = args.start_date
    stop_date = args.stop_date
    time_zone = args.time_zone

    # creates output dir, if it doesn't exist
    Path(out_dir).mkdir(parents=True, exist_ok=True)

    # Save log file
    sys.stdout = open(f"{out_dir}/logs_{start_date}_{stop_date}.txt", "a")

    obs_time, obs_unix, obs_unix_end = obs_times(time_zone, start_date, stop_date)

    # Lets make the a json file for each 30 min observation, with an empty list
    data = []
    for i in range(len(obs_time)):
        write_json(data, filename=f"{obs_time[i]}.json", out_dir=out_dir)

    # Finds all sat ephem json files, and loops over them
    for json_path in list(Path(json_dir).glob("*.json")):

        with open(json_path) as ephem:
            print(json_path)
            sat_ephem = json.load(ephem)

            # Extract data from json dictionary
            t_array = sat_ephem["time_array"]
            s_alt = sat_ephem["sat_alt"]
            s_az = sat_ephem["sat_az"]
            s_id = sat_ephem["sat_id"]

            # here, we're looping over each satellite pass with a single sat ephem file
            # to check which observation window it falls in

            for pass_idx in range(len(t_array)):

                # Don't consider passes with less than 4 samples (~ 1 min = 3*20s)
                if len(t_array[pass_idx]) > 3:
                    time_interp, sat_alt, sat_az = interp_ephem(
                        t_array[pass_idx],
                        s_alt[pass_idx],
                        s_az[pass_idx],
                        interp_type,
                        interp_freq,
                    )

                    # Loop over every sat pass, within the sat ephem file
                    for obs_int in range(len(obs_unix)):

                        sat_ephem = {}
                        sat_ephem["sat_id"] = [s_id[0]]
                        sat_ephem["time_array"] = []
                        sat_ephem["sat_alt"] = []
                        sat_ephem["sat_az"] = []

                        # Case I: Satpass occurs completely within the 30min observation
                        if (
                            obs_unix[obs_int] < time_interp[0]
                            and obs_unix_end[obs_int] > time_interp[-1]
                        ):

                            # append the whole pass to the dict
                            sat_ephem["time_array"].extend(time_interp)
                            sat_ephem["sat_alt"].extend(sat_alt)
                            sat_ephem["sat_az"].extend(sat_az)
                            # print(f'{pass_idx}: I.   {obs_time[obs_int]}')

                        # Case II: Satpass begins before the obs, but ends within it
                        elif (
                            obs_unix[obs_int] > time_interp[0]
                            and obs_unix[obs_int] < time_interp[-1]
                            and obs_unix_end[obs_int] > time_interp[-1]
                        ):

                            # find index of time_interp == obs_unix
                            start_idx = (
                                np.where(np.asarray(time_interp) == obs_unix[obs_int])
                            )[0][0]

                            # append the end of the pass which is within the obs
                            sat_ephem["time_array"].extend(time_interp[start_idx:])
                            sat_ephem["sat_alt"].extend(sat_alt[start_idx:])
                            sat_ephem["sat_az"].extend(sat_az[start_idx:])

                            # print(f'{pass_idx}: II.  {obs_time[obs_int]}')

                        # Case III: Satpass begins within the obs, but ends after it
                        elif (
                            obs_unix_end[obs_int] > time_interp[0]
                            and obs_unix_end[obs_int] < time_interp[-1]
                            and obs_unix[obs_int] < time_interp[0]
                        ):

                            # find index of time_interp == obs_unix_end
                            stop_idx = (
                                np.where(
                                    np.asarray(time_interp) == obs_unix_end[obs_int]
                                )
                            )[0][0]

                            # append the end of the pass which is within the obs
                            sat_ephem["time_array"].extend(time_interp[: stop_idx + 1])
                            sat_ephem["sat_alt"].extend(sat_alt[: stop_idx + 1])
                            sat_ephem["sat_az"].extend(sat_az[: stop_idx + 1])

                            # print(f'{pass_idx}: III. {obs_time[obs_int]}')

                        # doesn't create json if there are no satellite passes within it
                        if sat_ephem["time_array"] != []:

                            print(f"Satellite {s_id[0]} in {obs_time[obs_int]}")

                            # open the relevant json file and loads contents to 'data_json'
                            with open(
                                f"{out_dir}/{obs_time[obs_int]}.json"
                            ) as json_file:
                                data_json = json.load(json_file)

                                # append new satpass ephem data to data_json
                                data_json.append(sat_ephem)

                                # write the combined data back to the original file
                                write_json(
                                    data_json,
                                    filename=f"{obs_time[obs_int]}.json",
                                    out_dir=out_dir,
                                )

                                # clear data_json
                                data_json = []



"""
=========
Waterfall
=========

Saves a single waterfall plot to ``./embers_out/condition_data``

"""

import argparse
import pkg_resources
from pathlib import Path
from embers.condition_data.rf_data import single_waterfall


_parser = argparse.ArgumentParser(
    description="""
    Saved a single waterfall plot
    """
)

_parser.add_argument(
    "--rf_file", metavar="\b", default="", help="Path to raw rf data file"
)
_parser.add_argument(
    "--out_dir",
    metavar="\b",
    default="embers_out/condition_data",
    help="Dir where colormap sample plot is saved. Default=./embers_out/condition_data",
)

_args = _parser.parse_args()
_rf_file = _args.rf_file
_out_dir = _args.out_dir

# if no input file provided, use sample package data
if _rf_file == "":
    print("--------------------------------------------------")
    print("No input data provided, using packaged sample data")
    print(">>> waterfall --help, for more options")
    print("--------------------------------------------------")
    _rf_file = pkg_resources.resource_filename(
        "embers.kindle", "data/rf_data/S06XX/2019-10-10/S06XX_2019-10-10-02:30.txt"
    )


def main():
    """Execute waterfall from terminal."""

    _rf_name = Path(_rf_file).stem
    print(f"Waterfall plot saved to ./{_out_dir}/{_rf_name}.png")
    single_waterfall(_rf_file, _out_dir)



import sys
import numpy as np
import healpy as hp
from null_test import rotate
from scipy.stats import median_absolute_deviation as mad

sys.path.append("../sat_ephemeris")
from sat_ids import norad_ids


def good_maps(raw_tile):
    """Here, we extract "good satellie" data from the tile_maps_raw 
    generated by tile_maps.py. We also normalize the maps by
    the reference beam"""

    f_name, _ = raw_tile.name.split(".")
    tile, ref, _, _ = f_name.split("_")

    # Load reference FEE model
    # Rotate the fee models by -pi/2 to move model from spherical (E=0) to Alt/Az (N=0)
    ref_fee_model = np.load(ref_model, allow_pickle=True)
    if "XX" in tile:
        ref_fee = ref_fee_model["XX"]
        rotated_fee = rotate(nside, angle=-(1 * np.pi) / 2.0, healpix_array=ref_fee)
    else:
        ref_fee = ref_fee_model["YY"]
        rotated_fee = rotate(nside, angle=-(1 * np.pi) / 2.0, healpix_array=ref_fee)

    # load data from map .npz file
    tile_data = np.load(raw_tile, allow_pickle=True)
    tile_data = {key: tile_data[key].item() for key in tile_data}
    mwa_map = tile_data["mwa_map"]

    tile_maps_norm = {p: [] for p in pointings}
    mwa_maps_norm = {p: [] for p in pointings}

    for p in pointings:

        # mwa map
        mwa_map_good = [[] for pixel in range(hp.nside2npix(nside))]

        for sat in good_sats:

            for pix in range(hp.nside2npix(nside)):

                mwa_map_good[pix].extend(mwa_map[p][sat][pix])

        mwa_maps_norm[p].extend(mwa_map_good)

    # Save map arrays to npz file
    np.savez_compressed(f"{out_dir}/{tile}_{ref}_tile_maps.npz", **mwa_maps_norm)


if __name__ == "__main__":

    import argparse
    from pathlib import Path
    import concurrent.futures

    parser = argparse.ArgumentParser(
        description="""
        Extract the data from a list of 'good satellites' from the raw tile maps made my tile_maps.py. 
        Divide the tile/ref beams and apply the FEE ref model.
        """
    )

    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/tile_maps/tile_maps_norm/",
        help="Output directory. Default=./../../outputs/tile_maps/tile_maps_norm/",
    )
    parser.add_argument(
        "--map_dir",
        metavar="\b",
        default="./../../outputs/tile_maps/tile_maps_raw",
        help="Raw Map directory. Default=./../../outputs/tile_maps/tile_maps_raw",
    )
    parser.add_argument(
        "--ref_model",
        metavar="\b",
        default="../../outputs/reproject_ref/ref_dipole_models.npz",
        help="Healpix reference FEE model file. default=../../outputs/reproject_ref/ref_dipole_models.npz",
    )
    parser.add_argument(
        "--nside",
        metavar="\b",
        type=int,
        default=32,
        help="Healpix Nside. Default = 32",
    )

    args = parser.parse_args()

    out_dir = Path(args.out_dir)
    map_dir = Path(args.map_dir)
    ref_model = Path(args.ref_model)
    nside = args.nside

    # make outpud dir
    out_dir.mkdir(parents=True, exist_ok=True)

    # Good sats from which to make plots
    good_sats = [
        25338,
        25982,
        25984,
        25985,
        28654,
        40086,
        40087,
        40091,
        41179,
        41180,
        41182,
        41183,
        41184,
        41185,
        41187,
        41188,
        41189,
        44387,
    ]

    # use all satellites
    # good_sats = list(norad_ids.values())

    # list of beam pointings
    pointings = ["0", "2", "4", "41"]

    # list of all raw tile maps
    map_files = [item for item in map_dir.glob("*.npz")]

    # Parallization magic happens here
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = executor.map(good_maps, map_files)

#    good_maps(map_files[0])



import sys
import numpy as np
import healpy as hp
from mwa_pb.mwa_sweet_spots import all_grid_points
from scipy.stats import median_absolute_deviation as mad


def plot_healpix(data_map=None, sub=None, title=None, vmin=None, vmax=None, cmap=None):
    """Yeesh do some healpix magic to plot the thing"""

    # Healpix plotting script adapted from Dr. Jack Line's code
    # https://github.com/JLBLine/MWA_ORBCOMM

    # Disable cryptic healpy warnings. Can't figure out where they originate
    import warnings

    warnings.filterwarnings("ignore", category=RuntimeWarning)

    if vmin == None:
        if cmap == None:
            half_sky = hp.orthview(
                map=data_map,
                coord="E",
                half_sky=True,
                xsize=400,
                title=title,
                rot=(0, 90, 180),
                sub=sub,
                notext=True,
                return_projected_map=True,
            )
        else:
            half_sky = hp.orthview(
                map=data_map,
                coord="E",
                half_sky=True,
                xsize=400,
                title=title,
                rot=(0, 90, 180),
                sub=sub,
                cmap=cmap,
                notext=True,
                return_projected_map=True,
            )
    else:
        if cmap == None:
            half_sky = hp.orthview(
                map=data_map,
                coord="E",
                half_sky=True,
                xsize=400,
                rot=(0, 90, 180),
                title=title,
                sub=sub,
                min=vmin,
                max=vmax,
                notext=True,
                return_projected_map=True,
            )
        else:
            half_sky = hp.orthview(
                map=data_map,
                coord="E",
                half_sky=True,
                xsize=400,
                rot=(0, 90, 180),
                title=title,
                sub=sub,
                min=vmin,
                max=vmax,
                cmap=cmap,
                notext=True,
                return_projected_map=True,
            )

    hp.graticule(dpar=10, coord="E", color="k", alpha=0.3, dmer=45)

    # Altitude grid
    hp.projtext(00.0 * (np.pi / 180.0), 225.0 * (np.pi / 180), "0", coord="E")
    hp.projtext(30.0 * (np.pi / 180.0), 225.0 * (np.pi / 180), "30", coord="E")
    hp.projtext(60.0 * (np.pi / 180.0), 225.0 * (np.pi / 180), "60", coord="E")

    # Azimuth grid
    hp.projtext(
        75.0 * (np.pi / 180.0),
        000.0 * (np.pi / 180.0),
        r"$0^\circ$",
        coord="E",
        color="w",
        fontsize=10,
        verticalalignment="top",
    )
    hp.projtext(
        75.0 * (np.pi / 180.0),
        090.0 * (np.pi / 180.0),
        r"$90^\circ$",
        coord="E",
        color="w",
        fontsize=10,
        horizontalalignment="right",
    )
    hp.projtext(
        75.0 * (np.pi / 180.0),
        180.0 * (np.pi / 180.0),
        r"$180^\circ$",
        coord="E",
        color="w",
        fontsize=10,
        verticalalignment="bottom",
    )
    hp.projtext(
        75.0 * (np.pi / 180.0),
        270.0 * (np.pi / 180.0),
        r"$270^\circ$",
        coord="E",
        color="w",
        fontsize=10,
        horizontalalignment="left",
    )

    # NSEW
    hp.projtext(
        90.0 * (np.pi / 180.0),
        000.0 * (np.pi / 180.0),
        r"$N  $",
        coord="E",
        color="k",
        fontsize=14,
        verticalalignment="bottom",
    )
    hp.projtext(
        90.0 * (np.pi / 180.0),
        090.0 * (np.pi / 180.0),
        r"$E  $",
        coord="E",
        color="k",
        fontsize=14,
        horizontalalignment="left",
    )
    hp.projtext(
        90.0 * (np.pi / 180.0),
        180.0 * (np.pi / 180.0),
        r"$S  $",
        coord="E",
        color="k",
        fontsize=14,
        verticalalignment="top",
    )
    hp.projtext(
        90.0 * (np.pi / 180.0),
        270.0 * (np.pi / 180.0),
        r"$W  $",
        coord="E",
        color="k",
        fontsize=14,
        horizontalalignment="right",
    )


def plt_good_maps(f):

    f_name, _ = f.name.split(".")
    tile, ref, _, _ = f_name.split("_")

    pointings = ["0", "2", "4", "41"]

    # load data from map .npz file
    tile_data = np.load(f, allow_pickle=True)

    for p in pointings:

        ## find the pointing center in radians
        # pointing_center_az = np.radians(all_grid_points[int(p)][1])
        # pointing_center_za = np.radians(all_grid_points[int(p)][3])
        #
        ## convert it to a healpix vector
        # pointing_vec = hp.ang2vec(pointing_center_za, pointing_center_az)

        ## find all healpix indices within 10 degrees of pointing center
        # ipix_disc = hp.query_disc(nside=nside, vec=pointing_vec, radius=np.radians(10))

        # healpix meadian map
        tile_map_med = np.asarray(
            [(np.nanmedian(i) if i != [] else np.nan) for i in tile_data[p]]
        )

        ## find the max value within 10 degrees of pointing center
        # ipix_max = np.nanmax(tile_map_med[ipix_disc])
        #
        ## scale map such that the above max is set to 0dB
        # tile_map_scaled = np.asarray([(i - ipix_max) for i in tile_map_med])

        fig = plt.figure(figsize=(8, 10))
        fig.suptitle(f"Good Map: {tile}/{ref} @ {p}", fontsize=16)
        plot_healpix(data_map=tile_map_med, sub=(1, 1, 1), cmap=jade, vmin=-50, vmax=0)
        plt.savefig(
            f"{out_dir}/{p}/tile_maps/{tile}_{ref}_{p}_good_map.png",
            bbox_inches="tight",
        )
        plt.close()

        # Plot MAD
        tile_map_mad = []
        for j in tile_data[p]:
            if j != []:
                j = np.asarray(j)
                j = j[~np.isnan(j)]
                tile_map_mad.append(mad(j))
            else:
                tile_map_mad.append(np.nan)

        vmin = np.nanmin(tile_map_mad)
        vmax = np.nanmax(tile_map_mad)

        fig = plt.figure(figsize=(8, 10))
        fig.suptitle(f"Good Map MAD: {tile}/{ref} @ {p}", fontsize=16)
        plot_healpix(
            data_map=np.asarray(tile_map_mad),
            sub=(1, 1, 1),
            cmap=jade,
            vmin=vmin,
            vmax=vmax,
        )
        plt.savefig(
            f"{out_dir}/{p}/tile_errors/{tile}_{ref}_{p}_good_map_errors.png",
            bbox_inches="tight",
        )
        plt.close()

        # Plot counts in pix
        tile_map_counts = [len(np.array(i)[~np.isnan(i)]) for i in tile_data[p]]

        fig = plt.figure(figsize=(8, 10))
        fig.suptitle(f"Good Map Counts: {tile}/{ref} @ {p}", fontsize=16)
        plot_healpix(
            data_map=np.asarray(tile_map_counts),
            sub=(1, 1, 1),
            cmap=jade,
            vmin=0,
            vmax=80,
        )
        plt.savefig(
            f"{out_dir}/{p}/tile_counts/{tile}_{ref}_{p}_good_map_counts.png",
            bbox_inches="tight",
        )
        plt.close()


if __name__ == "__main__":

    import argparse
    import matplotlib

    matplotlib.use("Agg")
    from pathlib import Path
    import concurrent.futures
    import matplotlib.pyplot as plt

    import sys

    sys.path.append("../decode_rf_data")
    from colormap import spectral, jade, kelp

    # Custom spectral colormap
    jade, _ = jade()

    parser = argparse.ArgumentParser(
        description="""
        Plot healpix maps of normalized tile maps
        """
    )

    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/tile_maps/good_maps",
        help="Output directory. Default=./../../outputs/tile_maps/good_maps",
    )
    parser.add_argument(
        "--map_dir",
        metavar="\b",
        default="./../../outputs/tile_maps/tile_maps_norm",
        help="Output directory. Default=./../../outputs/tile_maps/tile_maps_norm",
    )
    parser.add_argument(
        "--nside",
        metavar="\b",
        type=int,
        default=32,
        help="Healpix Nside. Default = 32",
    )

    args = parser.parse_args()

    out_dir = Path(args.out_dir)
    map_dir = Path(args.map_dir)
    nside = args.nside

    pointings = ["0", "2", "4", "41"]

    map_files = [item for item in map_dir.glob("*.npz")]

    # Create output directory tree
    for p in pointings:
        Path(f"{out_dir}/{p}/tile_maps/").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/{p}/tile_counts/").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/{p}/tile_errors/").mkdir(parents=True, exist_ok=True)

    # Parallization magic happens here
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = executor.map(plt_good_maps, map_files)



"""
RF Data
=======

A set of tools to decode raw rf data recored by
RF Explorers and visualise waterfall plots.

"""

import re
import time
import numpy as np
from pathlib import Path
from itertools import product
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
from embers.condition_data.colormaps import spectral

_spec, _ = spectral()


def read_data(rf_file=None):
    """
    Convert raw rf binary data into a power array and extracts a list of UNIX timestamps.

    Parameters
    ----------
    :param rf_file: path to a rf data file
    :type rf_file: str

    Returns
    -------
    :returns:
        - power - array of rf powers in dBm
        - times - array of UNIX times

    :rtype: (float, numpy.array(float))

    """

    with open(rf_file, "rb") as f:
        next(f)
        lines = f.readlines()

        times = []
        data_lines = []

        for line in lines:
            time, data = line.split("$Sp".encode())
            times.append(time.decode())

            # List converts bytes to list of bytes
            # The last two charachters are excluded - Newline char
            data_lines.append(list(data[:-2]))

        # The (-1/2) converts an unsigned byte to a real value
        power = np.single(np.asarray(data_lines) * (-1 / 2))
        times = np.double(np.asarray(times))

        return (power, times)


def tile_names():
    """List of reference and MWA tile names.

    Returns
    -------
    :return: list of tile names
    :rtype: list[str]

    """

    tiles = [
        "rf0XX",
        "rf0YY",
        "rf1XX",
        "rf1YY",
        "S06XX",
        "S06YY",
        "S07XX",
        "S07YY",
        "S08XX",
        "S08YY",
        "S09XX",
        "S09YY",
        "S10XX",
        "S10YY",
        "S12XX",
        "S12YY",
        "S29XX",
        "S29YY",
        "S30XX",
        "S30YY",
        "S31XX",
        "S31YY",
        "S32XX",
        "S32YY",
        "S33XX",
        "S33YY",
        "S34XX",
        "S34YY",
        "S35XX",
        "S35YY",
        "S36XX",
        "S36YY",
    ]

    return tiles


def tile_pairs(tiles):
    """
    Create a list of all possible AUT ref antenna pairs
    
    Parameters
    ----------
    :param tiles: list of tile names
    :type tiles: list[str]

    Returns
    -------
    :return: list of possible tile pairs
    :rtype: list[list, str]

    """

    tiles = tile_names()
    refs = tiles[:4]
    AUTS = tiles[4:]

    # Split the list into XX and YY lists
    refs_XX = refs[::2]
    refs_YY = refs[1::2]

    AUTS_XX = AUTS[::2]
    AUTS_YY = AUTS[1::2]

    # Create a list of pairs of tiles to be aligned.
    # Essentially all possible combinations of Refs, AUTS
    tile_pairs = []
    for pair in list(product(refs_XX, AUTS_XX)):
        tile_pairs.append(pair)

    for pair in list(product(refs_YY, AUTS_YY)):
        tile_pairs.append(pair)

    return tile_pairs


def time_tree(start_date, stop_date):
    """Split a date interval into 30 min observation chunks.

    This is used to travers the following directory tree.
    The data_root dir contains a sub-dir for every tile in
    :func:`~embers.condition_data.rf_data.tile_names` within
    which are dirs for every day, containg raw rf data files
    recorded every 30 minutes

    .. code-block:: text
    
        data_root
        ├── tile1
        │   └── YYYY-MM-DD
        │       └── tile1_YYYY-MM-DD-HH:MM.txt
        └── tile2 
            └── YYYY-MM-DD
                └── tile2_YYYY-MM-DD-HH:MM.txt

    Parameters
    ----------
    :param start_date: Start date in ``YYYY-MM-DD`` format
    :type start_date: str
    :param stop_date: Stop date is inclusive, in ``YYYY-MM-DD`` format. 
    :type stop_date: str

    Returns
    -------
    :returns:
        - dates - list of dates between `start_date` and `stop_date`
        - time_stamps - list of list of 30 minute observation in each day

    :rtype: list[str]


    """

    t_start = datetime.strptime(start_date, "%Y-%m-%d")
    t_stop = datetime.strptime(stop_date, "%Y-%m-%d")
    n_days = (t_stop - t_start).days

    dates = []
    time_stamps = []

    # Every Day
    for i in range(n_days + 1):
        day = t_start + timedelta(days=i)
        date = day.strftime("%Y-%m-%d")
        dates.append(date)
        d_t = []

        # Every 30 min in the day
        for j in range(48):
            t_delta = datetime.strptime(date, "%Y-%m-%d") + timedelta(minutes=30 * j)
            d_time = t_delta.strftime("%Y-%m-%d-%H:%M")
            d_t.append(d_time)

        time_stamps.append(d_t)

    return (dates, time_stamps)


def plt_waterfall(power, times, name):
    """
    Create waterfall `.pyplot.plot` object.

    waterfall created using parameters ``power``, ``times``
    from :func:`~embers.condition_data.rf_data.read_data`. 
    Default unix times are converted to a human readable
    *HH:MM* format.

    Parameters
    ----------
    :param power: `numpy.ndarray` object from :func:`~embers.condition_data.rf_data.read_data`
    :param times: `numpy.ndarray` object from :func:`~embers.condition_data.rf_data.read_data`
    :param name: name of rf data identifier to add to plot title

    Returns
    -------
        - plt - `.pyplot.plot` object

    """

    # setting dynamic range of waterfall to be 30 dB above the median
    power_median = np.median(power)
    image = power - power_median
    vmin = 0
    vmax = 30

    plt.style.use("dark_background")
    fig = plt.figure(figsize=(7, 10))
    ax = fig.add_axes([0.12, 0.1, 0.72, 0.85])
    im = ax.imshow(image, vmin=vmin, vmax=vmax, interpolation="none", cmap=_spec)
    cax = fig.add_axes([0.88, 0.1, 0.03, 0.85])
    fig.colorbar(im, cax=cax)
    ax.set_aspect("auto")
    ax.set_title(f"Waterfall plot: {name}")

    # Number of time steps on y-axis
    number_t = 5
    t_step = int(len(times) / (number_t - 1))
    times = list(times)
    times = times[::t_step]

    # Convert UNIX time to local HH:MM time
    t_tz = []
    for i in range(len(times)):
        perth_t = float(times[i]) + 28800  # 28800=+8GMT @ PERTH
        hms = time.strftime("%H:%M", time.gmtime(perth_t))
        t_tz.append(hms)

    # Frequency: x-axis
    start_freq = 137.15
    stop_freq = 138.55

    # X-axis stuff
    x_ax = image.shape[1]
    freqs = np.arange(start_freq, stop_freq, 0.25)
    x_ticks = np.arange(0, x_ax, (0.25 / 0.0125))  # .0125MHz/ch
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(freqs)
    ax.set_xlabel("Freq [MHz]")

    # Y-axis stuff
    y_ax = image.shape[0]
    y_ticks = np.arange(0, y_ax, t_step)
    ax.set_yticks(y_ticks)
    ax.set_yticklabels(t_tz)
    ax.set_ylabel("Time [HH:MM]")

    return plt


def single_waterfall(rf_file, out_dir):
    """Save a waterfall plot from rf data file.
    
    Parameters
    ----------
    :param rf_file: path to a rf data file
    :type rf_file: str
    :param out_dir: path to output directory
    :type out_dir: str

    Returns
    -------
    :return: waterfall plot saved by `matplotlib.pyplot.savefig`

    """

    rf_name = Path(rf_file).stem

    power, times = read_data(rf_file)
    plt = plt_waterfall(power, times, rf_name)

    # Make out_dir if it doesn't exist
    Path(out_dir).mkdir(parents=True, exist_ok=True)
    plt.savefig(f"{out_dir}/{rf_name}.png")
    plt.close()


def batch_waterfall(tile, time_stamp, data_dir, out_dir):
    """Save a waterfall plot for a batch of rf data files.

    Parameters
    ----------
    :param tile: tile name 
    :type tile: str
    :param time_stamp: when the rf observation began 
    :type time_stamp: str [*YYYY-MM-DD-HH:MM*]
    :param data_dir: path to root of data directory
    :type data_dir: str
    :param out_dir: path to output directory
    :type out_dir: str

    Returns
    -------
    :return: waterfall plot saved by `matplotlib.pyplot.savefig`
    
    Exeptions
    ---------
    :raises FileNotFoundError: input file does not exist

    """

    rf_name = f"{tile}_{time_stamp}"
    date = re.search(r"\d{4}.\d{2}.\d{2}", time_stamp)[0]
    rf_path = Path(f"{data_dir}/{tile}/{date}/{rf_name}.txt")

    try:
        open(rf_path, "r")
    except FileNotFoundError as e:
        return e
    except Exception as e:
        return e

    power, times = read_data(rf_path)
    plt = plt_waterfall(power, times, rf_name)

    # Make out_dir if it doesn't exist
    save_dir = Path(f"{out_dir}/waterfalls/{date}/{time_stamp}")
    save_dir.mkdir(parents=True, exist_ok=True)

    plt.savefig(f"{save_dir}/{rf_name}.png")
    plt.close()

    return f"Waterfall plot saved to {save_dir}/{rf_name}.png"



# A dictionary of satellites which emit between 137.150 and 138.550 MHz
# ORBCOMM Communication satellites, NOAA & METEOR Weather satellites
# Satellite Name : NORAD CATALOGUE ID

norad_ids = {
    "ORBCOMM-X": 21576,
    "ORBCOMM FM 1": 23545,
    "ORBCOMM FM 2": 23546,
    "ORBCOMM FM 8": 25112,
    "ORBCOMM FM 10": 25113,
    "ORBCOMM FM 11": 25114,
    "ORBCOMM FM 12": 25115,
    "ORBCOMM FM 9": 25116,
    "ORBCOMM FM 5": 25117,
    "ORBCOMM FM 6": 25118,
    "ORBCOMM FM 7": 25119,
    "ORBCOMM FM 3": 25158,
    "ORBCOMM FM 4": 25159,
    "ORBCOMM FM 17": 25413,
    "ORBCOMM FM 18": 25414,
    "ORBCOMM FM 19": 25415,
    "ORBCOMM FM 20": 25416,
    "ORBCOMM FM 16": 25417,
    "ORBCOMM FM 15": 25418,
    "ORBCOMM FM 14": 25419,
    "ORBCOMM FM 13": 25420,
    "ORBCOMM FM 21": 25475,
    "ORBCOMM FM 22": 25476,
    "ORBCOMM FM 23": 25477,
    "ORBCOMM FM 24": 25478,
    "ORBCOMM FM 25": 25479,
    "ORBCOMM FM 26": 25480,
    "ORBCOMM FM 27": 25481,
    "ORBCOMM FM 28": 25482,
    "ORBCOMM FM 30": 25980,
    "ORBCOMM FM 31": 25981,
    "ORBCOMM FM 32": 25982,
    "ORBCOMM FM 33": 25983,
    "ORBCOMM FM 36": 25984,
    "ORBCOMM FM 35": 25985,
    "ORBCOMM FM 34": 25986,
    "ORBCOMM FM 38": 33060,
    "ORBCOMM FM 41": 33061,
    "ORBCOMM FM 29": 33062,
    "ORBCOMM FM 39": 33063,
    "ORBCOMM FM 37": 33064,
    "ORBCOMM FM 40": 33065,
    "ORBCOMM FM 109": 40086,
    "ORBCOMM FM 107": 40087,
    "ORBCOMM FM 106": 40088,
    "ORBCOMM FM 111": 40089,
    "ORBCOMM FM 104": 40090,
    "ORBCOMM FM 103": 40091,
    "ORBCOMM FM 114": 41179,
    "ORBCOMM FM 119": 41180,
    "ORBCOMM FM 105": 41181,
    "ORBCOMM FM 110": 41182,
    "ORBCOMM FM 118": 41183,
    "ORBCOMM FM 112": 41184,
    "ORBCOMM FM 113": 41185,
    "ORBCOMM FM 115": 41186,
    "ORBCOMM FM 108": 41187,
    "ORBCOMM FM 117": 41188,
    "ORBCOMM FM 116": 41189,
    "ORBCOMM FM 16 DEB 1": 44018,
    "ORBCOMM FM 16 DEB 2": 44019,
    "ORBCOMM FM 16 DEB 3": 44020,  # [NO CURRENT TLEs FOUND!]
    "ORBCOMM FM 16 DEB 4": 44022,
    "ORBCOMM FM 16 DEB 5": 44023,
    "ORBCOMM FM 16 DEB 6": 44024,
    "ORBCOMM FM 16 DEB 7": 44025,  # [NO CURRENT TLEs FOUND!]
    "ORBCOMM FM 16 DEB 8": 44026,
    "ORBCOMM FM 16 DEB 9": 44027,
    "ORBCOMM FM 16 DEB 10": 44028,
    "NOAA 18": 28654,
    "NOAA 15": 25338,
    "Meteor M2": 40069,
    "Meteor M2-2": 44387,
}

if __name__ == "__main__":
    for ids in norad_ids.values():
        print(ids)



import sys
import json
import argparse
import matplotlib
import numpy as np
import healpy as hp

matplotlib.use("Agg")
from pathlib import Path
import concurrent.futures
import scipy.optimize as opt
from null_test import rotate
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.stats import chisquare
from scipy.stats import median_absolute_deviation as mad
from scipy.interpolate import make_interp_spline, BSpline
import scipy.optimize as opt
import numpy.polynomial.polynomial as poly

sys.path.append("../sat_ephemeris")
from sat_ids import norad_ids

sys.path.append("../sat_channels")
from sat_channels import time_tree, time_filter

# Custom spectral colormap
sys.path.append("../decode_rf_data")
from rf_data import tile_names
from colormap import spectral

cmap = spectral()


def check_pointing(timestamp, point_0, point_2, point_4, point_41):
    """Check if timestamp is at pointing 0, 2, 4, 41"""
    if timestamp in point_0:
        point = 0
    elif timestamp in point_2:
        point = 2
    elif timestamp in point_4:
        point = 4
    else:
        point = 41

    return point


def read_aligned(ali_file=None):
    """Read aligned data npz file"""
    paired_data = np.load(ali_file, allow_pickle=True)

    ref_p = paired_data["ref_p_aligned"]
    tile_p = paired_data["tile_p_aligned"]
    times = paired_data["time_array"]

    return [ref_p, tile_p, times]


def noise_floor(sat_thresh, noi_thresh, data=None):
    """Computes the noise floor of the data """

    # compute the standard deviation of data, and use it to identify occupied channels
    σ = np.std(data)

    # Any channel with a max power >= σ has a satellite
    sat_cut = sat_thresh * σ
    chans_pow_max = np.amax(data, axis=0)

    # Exclude the channels with sats, to only have noise data
    noise_chans = np.where(chans_pow_max < sat_cut)[0]
    noise_data = data[:, noise_chans]

    # noise median, noise mad, noise threshold = μ + 3*σ
    μ_noise = np.median(noise_data)
    σ_noise = mad(noise_data, axis=None)
    noise_threshold = μ_noise + noi_thresh * σ_noise

    return (data, noise_threshold)


# chisquared minimization to best fit map to data
def fit_gain(map_data=None, fee=None):
    """Fit the beam model to the measured data using
    chisquared minimization"""

    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    fee = fee[~bad_values]

    def chisqfunc(gain):
        model = fee + gain
        chisq = sum((map_data - model) ** 2)
        return chisq

    x0 = np.array([0])

    result = opt.minimize(chisqfunc, x0)

    return result.x


def fit_test(map_data=None, fee=None):
    """chi-squared test for fit between
    model and data"""

    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    fee = fee[~bad_values]

    # map_data = np.asarray(map_data) + 120
    # fee = np.asarray(fee) + 120
    map_data = np.asarray(map_data) - np.nanmin(fee) + 20
    fee = np.asarray(fee) - np.nanmin(fee) + 20

    _, pvalue = chisquare(map_data, f_exp=fee)

    return pvalue


def poly_fit(x, y, order):
    """Fit polynominal of order to data"""

    x = np.asarray(x)
    y = np.asarray(y)

    coefs = poly.polyfit(x, y, order)
    fit = poly.polyval(x, coefs)
    return fit


def plt_fee_fit(t, mwa_fee_pass, mwa_pass_fit, out_dir, point, timestamp, sat, p_val):

    pval = fit_test(map_data=mwa_pass_fit, fee=mwa_fee_pass)

    plt.style.use("seaborn")

    # fig = plt.figure(figsize=(3.6,2.4))
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)
    ax1.set_title(f"P_val: {p_val}")

    t = np.array(t)
    mwa_fee_pass = [x for _, x in sorted(zip(t, mwa_fee_pass))]
    mwa_pass_fit = [x for _, x in sorted(zip(t, mwa_pass_fit))]
    t = sorted(t)
    t = (t - t[0]) / 60

    # t_new = np.linspace(t.min(), t.max(), 49)
    # spl = make_interp_spline(t, mwa_fee_pass, k=3)  # type: BSpline
    # mwa_fee_pass = spl(t_new)
    #
    # spl = make_interp_spline(t, mwa_pass_fit, k=3)  # type: BSpline
    # mwa_pass_fit = spl(t_new)

    # ax1.plot(t_new, mwa_fee_pass, color='#7da87b', alpha=0.9, lw=2, label=r"$B_{FEE}$")
    # ax1.plot(t_new, mwa_pass_fit, color='#c70039', alpha=0.9, lw=2, label=r"$B_{dist}$")
    ax1.scatter(
        t, mwa_fee_pass, color="#7da87b", alpha=0.9, marker=".", s=36, label="FEE model"
    )
    ax1.scatter(
        t, mwa_pass_fit, color="#c70039", alpha=0.9, marker=".", s=36, label="RF data"
    )
    ax1.set_ylabel("Power [dBm]")
    ax1.set_ylim([-84, -16])
    ax1.set_xticklabels([])

    leg = ax1.legend(loc="upper left", frameon=True, markerscale=2, handlelength=1)
    leg.get_frame().set_facecolor("grey")
    leg.get_frame().set_alpha(0.4)
    for l in leg.legendHandles:
        l.set_alpha(1)

    delta_p = np.array(mwa_fee_pass) - np.array(mwa_pass_fit)

    divider = make_axes_locatable(ax1)
    dax = divider.append_axes("bottom", size="40%", pad=0.10)

    dax.scatter(
        t, delta_p, color="#27296d", alpha=0.9, marker=".", s=36, label="Residuals"
    )
    leg = dax.legend(loc="lower left", frameon=True, markerscale=2, handlelength=1)
    leg.get_frame().set_facecolor("grey")
    leg.get_frame().set_alpha(0.4)
    for l in leg.legendHandles:
        l.set_alpha(1)
    # dax.scatter(t, delta_p, marker='.', s=21,alpha=0.7, color='#27296d')
    dax.set_xlabel("Times [min]")
    # dax.plot(zen_angle, pow_fit, linewidth=1.4, alpha=0.9, color='#ff8264')
    # dax.set_ylim([-21,12])
    dax.set_yticks([-20, 0, 10])
    # dax.set_xticklabels([])
    # dax.set_ylim([-5,5])
    dax.set_ylabel(r"$\Delta$P [dBm]")

    plt.tight_layout()
    plt.savefig(f"{out_dir}/{timestamp}_{sat}.png", bbox_inches="tight")
    plt.close()


def power_ephem(
    ref, tile, ali_file, chrono_file, sat_id, sat_chan, point, pow_thresh, timestamp
):

    """Create power, alt, az arrays at constant cadence"""

    # Read .npz aligned file
    ref_p, tile_p, times = read_aligned(ali_file=ali_file)

    # Scale noise floor to zero and determine noise threshold
    ref_p, ref_noise = noise_floor(sat_thresh, noi_thresh, ref_p)
    tile_p, tile_noise = noise_floor(sat_thresh, noi_thresh, tile_p)

    with open(chrono_file) as chrono:
        chrono_ephem = json.load(chrono)

        norad_list = [chrono_ephem[s]["sat_id"][0] for s in range(len(chrono_ephem))]

        norad_index = norad_list.index(sat_id)

        norad_ephem = chrono_ephem[norad_index]

        rise_ephem = norad_ephem["time_array"][0]
        set_ephem = norad_ephem["time_array"][-1]

        intvl = time_filter(rise_ephem, set_ephem, np.asarray(times))

        if intvl != None:

            w_start, w_stop = intvl

            # Slice [crop] the ref/tile/times arrays to the times of sat pass and extract sat_chan
            ref_c = ref_p[w_start : w_stop + 1, sat_chan]
            tile_c = tile_p[w_start : w_stop + 1, sat_chan]
            times_c = times[w_start : w_stop + 1]

            alt = np.asarray(norad_ephem["sat_alt"])
            az = np.asarray(norad_ephem["sat_az"])

            if (np.nanmax(ref_c - ref_noise) >= pow_thresh) and (
                np.nanmax(tile_c - tile_noise) >= pow_thresh
            ):

                # Apply noise criteria. In the window, where are ref_power and tile power
                # above their respective thresholds?
                if np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0].size != 0:
                    good_ref = ref_c[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]
                    good_tile = tile_c[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]
                    good_alt = alt[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]
                    good_az = az[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]

                    return [good_ref, good_tile, good_alt, good_az, times_c]

                else:
                    return 0

            else:
                return 0

        else:
            return 0


def rfe_gain(tile_pair):

    # pass_data = []
    # pass_resi = []

    resi_gain = {}
    resi_gain["pass_data"] = []
    resi_gain["pass_resi"] = []

    ref, tile = tile_pair

    pointings = ["0", "2", "4", "41"]

    if plots == "True":
        Path(f"{out_dir}/fit_plots/").mkdir(parents=True, exist_ok=True)

    # Initialize an empty dictionary for tile data
    # The map is list of length 12288 of empty lists to append pixel values to
    # keep track of which satellites contributed which data
    tile_data = {
        "mwa_maps": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "ref_maps": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "tile_maps": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "sat_map": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "times": {p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings},
    }

    # Load reference FEE model
    # Rotate the fee models by -pi/2 to move model from spherical (E=0) to Alt/Az (N=0)
    ref_fee_model = np.load(ref_model, allow_pickle=True)

    fee_m = np.load(fee_map, allow_pickle=True)

    if "XX" in tile:
        ref_fee = ref_fee_model["XX"]
        rotated_fee = rotate(nside, angle=-(1 * np.pi) / 2.0, healpix_array=ref_fee)
    else:
        ref_fee = ref_fee_model["YY"]
        rotated_fee = rotate(nside, angle=-(1 * np.pi) / 2.0, healpix_array=ref_fee)

    for day in range(len(dates)):

        for window in range(len(date_time[day])):
            timestamp = date_time[day][window]

            # Check if at timestamp, reciever was pointed to 0,2,4 gridpointing
            if (
                (timestamp in point_0)
                or (timestamp in point_2)
                or (timestamp in point_4)
                or (timestamp in point_41)
            ):

                # pointing at timestamp
                point = check_pointing(timestamp, point_0, point_2, point_4, point_41)

                if point is 0:

                    if "XX" in tile:
                        mwa_fee = fee_m[str(point)][0]
                    else:
                        mwa_fee = fee_m[str(point)][1]

                    ali_file = Path(
                        f"{align_dir}/{dates[day]}/{timestamp}/{ref}_{tile}_{timestamp}_aligned.npz"
                    )

                    # check if file exists
                    if ali_file.is_file():

                        # Chrono and map Ephemeris file
                        chrono_file = Path(f"{chrono_dir}/{timestamp}.json")
                        channel_map = f"{map_dir}/{timestamp}.json"

                        with open(chrono_file) as chrono:
                            chrono_ephem = json.load(chrono)

                            if chrono_ephem != []:

                                norad_list = [
                                    chrono_ephem[s]["sat_id"][0]
                                    for s in range(len(chrono_ephem))
                                ]

                                if norad_list != []:

                                    with open(channel_map) as ch_map:
                                        chan_map = json.load(ch_map)

                                        chan_sat_ids = [
                                            int(i) for i in list(chan_map.keys())
                                        ]

                                        for sat in chan_sat_ids:

                                            chan = chan_map[f"{sat}"]

                                            sat_data = power_ephem(
                                                ref,
                                                tile,
                                                ali_file,
                                                chrono_file,
                                                sat,
                                                chan,
                                                point,
                                                pow_thresh,
                                                timestamp,
                                            )

                                            if sat_data != 0:

                                                (
                                                    ref_power,
                                                    tile_power,
                                                    alt,
                                                    az,
                                                    times,
                                                ) = sat_data

                                                # Altitude is in deg while az is in radians
                                                # convert alt to radians
                                                # za - zenith angle
                                                alt = np.radians(alt)
                                                za = np.pi / 2 - alt
                                                az = np.asarray(az)

                                                # Now convert to healpix coordinates
                                                # healpix_index = hp.ang2pix(nside,θ, ɸ)
                                                healpix_index = hp.ang2pix(
                                                    nside, za, az
                                                )

                                                # multiple data points fall within a single healpix pixel
                                                # find the unique pixels
                                                u = np.unique(healpix_index)

                                                ref_pass = np.array(
                                                    [
                                                        np.nanmean(
                                                            ref_power[
                                                                np.where(
                                                                    healpix_index == i
                                                                )[0]
                                                            ]
                                                        )
                                                        for i in u
                                                    ]
                                                )
                                                tile_pass = np.array(
                                                    [
                                                        np.nanmean(
                                                            tile_power[
                                                                np.where(
                                                                    healpix_index == i
                                                                )[0]
                                                            ]
                                                        )
                                                        for i in u
                                                    ]
                                                )
                                                times_pass = np.array(
                                                    [
                                                        np.mean(
                                                            times[
                                                                np.where(
                                                                    healpix_index == i
                                                                )
                                                            ][0]
                                                        )
                                                        for i in u
                                                    ]
                                                )
                                                ref_fee_pass = np.array(
                                                    [rotated_fee[i] for i in u]
                                                )
                                                mwa_fee_pass = np.array(
                                                    [mwa_fee[i] for i in u]
                                                )

                                                # This is the magic. Equation [1] of the paper
                                                mwa_pass = (
                                                    np.array(tile_pass)
                                                    - np.array(ref_pass)
                                                    + np.array(ref_fee_pass)
                                                )

                                                # RFE distortion is seen in tile_pass when raw power is above -40dBm
                                                # fit the mwa_pass data to the tile_pass power level
                                                # Mask everything below -50dBm to fit distorted MWA and tile pass
                                                peak_filter = np.where(tile_pass >= -30)
                                                offset = fit_gain(
                                                    map_data=tile_pass[peak_filter],
                                                    fee=mwa_pass[peak_filter],
                                                )
                                                # This is a slice of the MWA beam, scaled back to the power level of the raw, distorted tile data
                                                mwa_pass = mwa_pass + offset[0]

                                                # Single multiplicative gain factor to fit MWA FEE beam slice down to tile pass power level
                                                # MWA pass Data above -50dBm masked out because it is distorted
                                                # Data below -60dBm maked out because FEE nulls are much deeper than the dynamic range of satellite passes
                                                dis_filter = np.where(mwa_pass <= -35)
                                                mwa_pass_fil = mwa_pass[dis_filter]
                                                mwa_fee_pass_fil = mwa_fee_pass[
                                                    dis_filter
                                                ]
                                                null_filter = np.where(
                                                    mwa_fee_pass_fil >= -55
                                                )

                                                offset = fit_gain(
                                                    map_data=mwa_pass_fil[null_filter],
                                                    fee=mwa_fee_pass_fil[null_filter],
                                                )
                                                mwa_fee_pass = mwa_fee_pass + offset
                                                mwa_pass_fit = mwa_pass

                                                # more than 30 non distorted samples
                                                if (
                                                    mwa_fee_pass[dis_filter][
                                                        null_filter
                                                    ].size
                                                    >= 30
                                                ):

                                                    # determine how well the data fits the model with chi-square
                                                    pval = fit_test(
                                                        map_data=mwa_pass_fit[
                                                            dis_filter
                                                        ][null_filter],
                                                        fee=mwa_fee_pass[dis_filter][
                                                            null_filter
                                                        ],
                                                    )

                                                    # a goodness of fit threshold
                                                    if pval >= 0.8:

                                                        # consider residuals of sats which pass within 10 deg of zenith
                                                        # an hp index of 111 approx corresponds to a zenith angle of 10 degrees
                                                        hp_10_deg = 111

                                                        if np.amin(u) <= hp_10_deg:

                                                            # only passes longer than 10 minutes
                                                            if (
                                                                np.amax(times_pass)
                                                                - np.amin(times_pass)
                                                            ) >= 600:

                                                                # residuals between scaled FEE and mwa pass
                                                                resi = (
                                                                    mwa_fee_pass
                                                                    - mwa_pass_fit
                                                                )
                                                                resi_gain[
                                                                    "pass_data"
                                                                ].extend(mwa_pass_fit)
                                                                resi_gain[
                                                                    "pass_resi"
                                                                ].extend(resi)

                                                                # Plot individual passes
                                                                if plots == "True":
                                                                    # if mwa_fee_pass.size !=0:
                                                                    #    if np.amax(mwa_fee_pass) >= -30:
                                                                    plt_fee_fit(
                                                                        times_pass,
                                                                        mwa_fee_pass,
                                                                        mwa_pass_fit,
                                                                        f"{out_dir}/fit_plots/",
                                                                        point,
                                                                        timestamp,
                                                                        sat,
                                                                        pval,
                                                                    )

    # Save gain residuals to json file
    with open(f"{out_dir}/{tile}_{ref}_gain_fit.json", "w") as outfile:
        json.dump(resi_gain, outfile, indent=4)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="""
        Project a sat pass onto a healpix map using ephemeris data
        """
    )

    parser.add_argument(
        "--align_dir",
        metavar="\b",
        default="../../outputs/align_data/",
        help="Dir where agligned data date is saved. Default:../../outputs/align_data",
    )

    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/tile_maps/rfe_gain/",
        help="Output directory. Default=./../../outputs/tile_maps/rfe_gain/",
    )

    parser.add_argument(
        "--obs_point",
        metavar="\b",
        default="../../outputs/beam_pointings/obs_pointings.json",
        help="Observation pointing lists. Default=../../outputs/beam_pointings/obs_pointings.json",
    )

    parser.add_argument(
        "--chrono_dir",
        metavar="\b",
        default="./../../outputs/sat_ephemeris/chrono_json",
        help="Output directory. Default=./../../outputs/sat_ephemeris/chrono_json/",
    )

    parser.add_argument(
        "--map_dir",
        metavar="\b",
        default="../../outputs/sat_channels/window_maps/",
        help="Satellite channel map. Default=../../outputs/sat_channels/window_maps/",
    )

    parser.add_argument(
        "--ref_model",
        metavar="\b",
        default="../../outputs/reproject_ref/ref_dipole_models.npz",
        help="Healpix reference FEE model file. default=../../outputs/reproject_ref/ref_dipole_models.npz",
    )

    parser.add_argument(
        "--fee_map",
        metavar="\b",
        default="../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz",
        help="Healpix FEE map of mwa tile. default=../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz",
    )

    parser.add_argument(
        "--fit_thresh",
        metavar="\b",
        default=0.9,
        help="Goodness of fit threshold. 0.9 seems to only reject obvious outliers",
    )
    parser.add_argument(
        "--noi_thresh",
        metavar="\b",
        type=int,
        default=3,
        help="Noise Threshold: Multiples of MAD. Default=3.",
    )
    parser.add_argument(
        "--sat_thresh",
        metavar="\b",
        type=int,
        default=1,
        help="σ threshold to detect sats Default=1.",
    )
    parser.add_argument(
        "--pow_thresh",
        metavar="\b",
        type=int,
        default=5,
        help="Power Threshold to detect sats. Default=10 dB.",
    )
    parser.add_argument(
        "--nside",
        metavar="\b",
        type=int,
        default=32,
        help="Healpix Nside. Default = 32",
    )
    parser.add_argument(
        "--plots",
        metavar="\b",
        default=False,
        help="If True, create a gazzillion plots for each sat pass. Default = False",
    )
    parser.add_argument(
        "--start_date",
        metavar="\b",
        help="Date from which to start aligning data. Ex: 2019-10-10",
    )
    parser.add_argument(
        "--stop_date",
        metavar="\b",
        help="Date until which to align data. Ex: 2019-10-11",
    )

    args = parser.parse_args()

    start_date = args.start_date
    stop_date = args.stop_date
    obs_point = args.obs_point
    noi_thresh = args.noi_thresh
    sat_thresh = args.sat_thresh
    pow_thresh = args.pow_thresh
    fit_thresh = args.fit_thresh
    nside = args.nside
    plots = args.plots
    ref_model = args.ref_model
    fee_map = args.fee_map

    align_dir = Path(args.align_dir)
    chrono_dir = Path(args.chrono_dir)
    out_dir = Path(args.out_dir)
    map_dir = Path(args.map_dir)

    # Tile names
    refs = tile_names()[:4]
    tiles = tile_names()[4:]

    # All relevant tile pairs
    tile_pairs = []
    for ref in refs:
        if "XX" in ref:
            for tile in [t for t in tiles if "XX" in t]:
                tile_pairs.append([ref, tile])
        else:
            for tile in [t for t in tiles if "YY" in t]:
                tile_pairs.append([ref, tile])

    # Read observation pointing list
    with open(obs_point) as point:
        obs_p = json.load(point)
        point_0 = obs_p["point_0"]
        point_2 = obs_p["point_2"]
        point_4 = obs_p["point_4"]
        point_41 = obs_p["point_41"]

    # dates: list of days
    # date_time = list of 30 min observation windows
    dates, date_time = time_tree(start_date, stop_date)

    # Save logs
    Path(out_dir).mkdir(parents=True, exist_ok=True)
    sys.stdout = open(f"{out_dir}/logs_{start_date}_{stop_date}.txt", "a")

    #    rfe_gain(tile_pairs[0])

    # Parallization magic happens here
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = executor.map(rfe_gain, tile_pairs)



import numpy as np
from align_data import savgol_interp
import concurrent.futures

from pathlib import Path
from itertools import repeat
from itertools import product
from datetime import datetime, timedelta

import sys

sys.path.append("../decode_rf_data")
import rf_data as rf


def save_aligned(
    ref,
    aut,
    ref_path,
    aut_path,
    savgol_window_1,
    savgol_window_2,
    polyorder,
    interp_type,
    interp_freq,
    out_dir,
    date,
    time_stamp,
):

    try:
        ref_file = f"{ref_path}/{ref}_{time_stamp}.txt"
        aut_file = f"{aut_path}/{aut}_{time_stamp}.txt"

        _, _, _, _, ref_p_aligned, tile_p_aligned, time_array = savgol_interp(
            ref_file,
            aut_file,
            savgol_window_1=savgol_window_1,
            savgol_window_2=savgol_window_2,
            polyorder=polyorder,
            interp_type=interp_type,
            interp_freq=interp_freq,
        )

        # creates output directory if it doesn't exist
        save_dir = Path(f"{out_dir}/{date}/{time_stamp}")
        save_dir.mkdir(parents=True, exist_ok=True)

        # Convert the power array to float32
        # Convert list of times to float64 (double)
        # Save as compressed npz file. Seems to drastically reduce size
        np.savez_compressed(
            f"{save_dir}/{ref}_{aut}_{time_stamp}_aligned.npz",
            ref_p_aligned=np.single(ref_p_aligned),
            tile_p_aligned=np.single(tile_p_aligned),
            time_array=np.double(time_array),
        )

        return f"Saving {ref}_{aut}_{time_stamp}_aligned.npz"

    except Exception:
        return f"Cound not save {ref}_{aut}_{time_stamp}_aligned.npz. Missing file"
        # This exception should only be raised if both files don't exist
        # In that case, the exception from rf_data.read_data() will be
        # displayed here.


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(
        description="""
        Time alignes and smoothes all data 
        between two date ranges. Saves data
        pairs in organised directory 
        structure as .npz files.
        """
    )

    parser.add_argument("--data_dir", metavar="\b", help="Dir where date is saved")
    parser.add_argument(
        "--start_date",
        metavar="\b",
        help="Date from which to start aligning data. Ex: 2019-10-10",
    )
    parser.add_argument(
        "--stop_date",
        metavar="\b",
        help="Date until which to align data. Ex: 2019-10-11",
    )
    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/align_data/",
        help="Output directory. Default=./../../outputs/align_data/",
    )
    parser.add_argument(
        "--savgol_window_1",
        metavar="\b",
        type=int,
        default=11,
        help="Length of savgol window. Must be odd. Default=11",
    )
    parser.add_argument(
        "--savgol_window_2",
        metavar="\b",
        type=int,
        default=15,
        help="Length of savgol window. Must be odd. Default=15",
    )
    parser.add_argument(
        "--polyorder",
        metavar="\b",
        type=int,
        default=2,
        help="Order of polynomial to fit to savgol window. Default=2",
    )
    parser.add_argument(
        "--interp_type",
        metavar="\b",
        default="cubic",
        help="Type of interpolation. Ex: cubic, linear, etc. Default=cubic",
    )
    parser.add_argument(
        "--interp_freq",
        metavar="\b",
        type=int,
        default=1,
        help="Frequency at which to resample smoothed data, in Hertz. Default=1",
    )

    args = parser.parse_args()

    data_dir = args.data_dir
    start_date = args.start_date
    stop_date = args.stop_date
    out_dir = args.out_dir
    savgol_window_1 = args.savgol_window_1
    savgol_window_2 = args.savgol_window_2
    polyorder = args.polyorder
    interp_type = args.interp_type
    interp_freq = args.interp_freq

    # Save logs
    Path(out_dir).mkdir(parents=True, exist_ok=True)
    sys.stdout = open(f"{out_dir}/logs_{start_date}_{stop_date}.txt", "a")

    # Import list of tile names from rf_data.py
    tiles = rf.tile_names()

    refs = tiles[:4]
    AUTS = tiles[4:]

    # Split the list into XX and YY lists
    refs_XX = refs[::2]
    refs_YY = refs[1::2]

    AUTS_XX = AUTS[::2]
    AUTS_YY = AUTS[1::2]

    # Create a list of pairs of tiles to be aligned.
    # Essentially all possible combinations of Refs, AUTS
    align_pairs = []
    for pair in list(product(refs_XX, AUTS_XX)):
        align_pairs.append(pair)

    for pair in list(product(refs_YY, AUTS_YY)):
        align_pairs.append(pair)

    # Time stuff to help traverse data dir tree.
    t_start = datetime.strptime(start_date, "%Y-%m-%d")
    t_stop = datetime.strptime(stop_date, "%Y-%m-%d")
    n_days = (t_stop - t_start).days

    dates = []
    date_time = []

    for i in range(n_days + 1):
        day = t_start + timedelta(days=i)
        date = day.strftime("%Y-%m-%d")
        dates.append(date)
        d_t = []

        for j in range(48):
            t_delta = datetime.strptime(date, "%Y-%m-%d") + timedelta(minutes=30 * j)
            d_time = t_delta.strftime("%Y-%m-%d-%H:%M")
            d_t.append(d_time)

        date_time.append(d_t)

    data_dir = Path(data_dir)
    out_dir = Path(out_dir)

    for pair in align_pairs:
        for d in range(len(dates)):
            ref = pair[0]
            aut = pair[1]

            ref_path = data_dir / ref / dates[d]
            aut_path = data_dir / aut / dates[d]

            # Parallization magic happens here
            with concurrent.futures.ProcessPoolExecutor() as executor:
                results = executor.map(
                    save_aligned,
                    repeat(ref),
                    repeat(aut),
                    repeat(ref_path),
                    repeat(aut_path),
                    repeat(savgol_window_1),
                    repeat(savgol_window_2),
                    repeat(polyorder),
                    repeat(interp_type),
                    repeat(interp_freq),
                    repeat(out_dir),
                    repeat(dates[d]),
                    date_time[d],
                )

            for result in results:
                print(result)



import json
import wget
import time
from pathlib import Path


def cerberus_metafit(gps_time, out_dir):
    cerberus_url = "http://ws.mwatelescope.org/metadata/fits?obs_id="
    print(f"Downloading {gps_time} metafits")
    wget.download(f"{cerberus_url}{gps_time}", f"{out_dir}/{gps_time}.metafits")


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(
        description="""
            Download metafits files to find
            dipole flagging of tiles
            """
    )

    parser.add_argument(
        "--obs_ids",
        metavar="\b",
        default="./../../outputs/beam_pointings/ultimate_pointing_times.json",
        help="Directory where json metadata files live. Default=./../../outputs/beam_pointings/ultimate_pointing_times.json",
    )

    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/beam_pointings/metafits/",
        help="Directory where json metadata files live. Default=./../../outputs/beam_pointings/metafits/",
    )

    args = parser.parse_args()
    obs_ids = args.obs_ids
    out_dir = Path(args.out_dir)

    out_dir.mkdir(parents=True, exist_ok=True)

    with open(obs_ids) as gps:
        gps_times = json.load(gps)["start_gps"]

    for i in gps_times[10::10]:
        cerberus_metafit(i, out_dir)
        print("\n")
        time.sleep(10)



import sys
import json
import argparse
import matplotlib
import numpy as np
import healpy as hp

matplotlib.use("Agg")
from pathlib import Path
import concurrent.futures
import scipy.optimize as opt
from null_test import rotate
import matplotlib.pyplot as plt
from scipy.stats import chisquare
from scipy.stats import median_absolute_deviation as mad

sys.path.append("../sat_ephemeris")
from sat_ids import norad_ids

sys.path.append("../sat_channels")
from sat_channels import time_tree, time_filter

# Custom spectral colormap
sys.path.append("../decode_rf_data")
from rf_data import tile_names
from colormap import spectral

cmap = spectral()


def check_pointing(timestamp, point_0, point_2, point_4, point_41):
    """Check if timestamp is at pointing 0, 2, 4, 41"""
    if timestamp in point_0:
        point = 0
    elif timestamp in point_2:
        point = 2
    elif timestamp in point_4:
        point = 4
    else:
        point = 41

    return point


def read_aligned(ali_file=None):
    """Read aligned data npz file"""
    paired_data = np.load(ali_file, allow_pickle=True)

    ref_p = paired_data["ref_p_aligned"]
    tile_p = paired_data["tile_p_aligned"]
    times = paired_data["time_array"]

    return [ref_p, tile_p, times]


def noise_floor(sat_thresh, noi_thresh, data=None):
    """Computes the noise floor of the data """

    # compute the standard deviation of data, and use it to identify occupied channels
    σ = np.std(data)

    # Any channel with a max power >= σ has a satellite
    sat_cut = sat_thresh * σ
    chans_pow_max = np.amax(data, axis=0)

    # Exclude the channels with sats, to only have noise data
    noise_chans = np.where(chans_pow_max < sat_cut)[0]
    noise_data = data[:, noise_chans]

    # noise median, noise mad, noise threshold = μ + 3*σ
    μ_noise = np.median(noise_data)
    σ_noise = mad(noise_data, axis=None)
    noise_threshold = μ_noise + noi_thresh * σ_noise

    return (data, noise_threshold)


# chisquared minimization to best fit map to data
def fit_gain(map_data=None, fee=None):
    """Fit the beam model to the measured data using
    chisquared minimization"""

    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    fee = fee[~bad_values]

    def chisqfunc(gain):
        model = fee + gain
        chisq = sum((map_data - model) ** 2)
        return chisq

    x0 = np.array([0])

    result = opt.minimize(chisqfunc, x0)

    return result.x


def fit_test(map_data=None, fee=None):
    """chi-squared test for fit between
    model and data"""

    bad_values = np.isnan(map_data)
    map_data = map_data[~bad_values]
    fee = fee[~bad_values]

    # map_data = np.asarray(map_data) + 120
    # fee = np.asarray(fee) + 120
    map_data = np.asarray(map_data) - np.nanmin(fee) + 20
    fee = np.asarray(fee) - np.nanmin(fee) + 20

    _, pvalue = chisquare(map_data, f_exp=fee)

    return pvalue


def plt_channel(
    out_dir,
    times,
    ref_c,
    tile_c,
    ref_noise,
    tile_noise,
    chan_num,
    sat_id,
    pointing,
    date,
    point,
):

    """Plot power in channel, with various thresholds
    
    Args:
        times:          Time array
        channel_power:  Power in channel
        chan_num:       Channel Number
        min_s:          Minimum signal in channel_power
        max_s:          Maximum signal in channel_power
        noise_threshold: Noise Threshold (n*MAD)
        sat_id:         Norad Cat ID
        date:           Date of observation
        """

    plt.style.use("seaborn")

    ylims = [min(ref_c), max(ref_c), min(tile_c), max(tile_c)]
    min_s = min(ylims)
    max_s = max(ylims)

    fig = plt.figure(figsize=(8, 6))
    fig.suptitle(f"Satellite [{sat_id}] Pass @ {date} in Channel: [{chan_num}]", y=0.98)

    ax1 = fig.add_subplot(2, 1, 1)
    ax1.plot(
        times,
        ref_c,
        linestyle="-",
        linewidth=2,
        alpha=1.0,
        color="#729d39",
        label="ref",
    )
    ax1.fill_between(times, y1=ref_c, y2=-120, color="#729d39", alpha=0.7)
    ax1.axhline(
        ref_noise,
        linestyle="-",
        linewidth=2,
        color="#36622b",
        label=f"Ref Cut: {ref_noise:.2f} dBm",
    )
    ax1.axhspan(-120, ref_noise, color="#36622b", alpha=0.7)
    ax1.set_ylim([min(ref_c) - 1, max(ref_c) + 1])
    ax1.set_ylabel("Power [dBm]")
    ax1.set_xticklabels([])

    leg = ax1.legend(frameon=True)
    leg.get_frame().set_facecolor("grey")
    leg.get_frame().set_alpha(0.2)
    for l in leg.legendHandles:
        l.set_alpha(1)

    ax2 = fig.add_subplot(2, 1, 2)
    ax2.plot(
        times,
        tile_c,
        linestyle="-",
        linewidth=2,
        alpha=1.0,
        color="#ff5656",
        label="tile",
    )
    ax2.fill_between(times, y1=tile_c, y2=-120, color="#ff5656", alpha=0.7)
    ax2.axhline(
        tile_noise,
        linestyle="-",
        linewidth=2,
        color="#970747",
        label=f"Tile Cut: {tile_noise:.2f} dBm",
    )
    ax2.axhspan(-120, tile_noise, color="#970747", alpha=0.7)
    ax2.set_ylim([min(tile_c) - 1, max(tile_c) + 1])
    ax2.set_ylabel("Power [dBm]")
    ax2.set_xlabel("Time [s]")

    leg = ax2.legend(frameon=True)
    leg.get_frame().set_facecolor("grey")
    leg.get_frame().set_alpha(0.2)
    for l in leg.legendHandles:
        l.set_alpha(1)

    plt.tight_layout()
    plt.subplots_adjust(top=0.94)
    plt.savefig(f"{out_dir}/{date}_{sat_id}_{chan_num}_channel.png")
    plt.close()


def plt_fee_fit(
    t, mwa_fee_pass, mwa_pass_fit, mwa_pass_fit_raw, out_dir, point, timestamp, sat
):

    pval = fit_test(map_data=mwa_pass_fit, fee=mwa_fee_pass)

    plt.style.use("seaborn")

    fig = plt.figure(figsize=(8, 6))
    ax1 = fig.add_subplot(1, 1, 1)

    ax1.scatter(
        t, mwa_fee_pass, color="#c70039", alpha=0.6, marker=".", label="fee_slice"
    )

    where_nan = np.isnan(mwa_pass_fit)

    ax1.scatter(
        t, mwa_pass_fit_raw, color="#7da87b", alpha=0.9, marker=".", label="RFE raw"
    )
    ax1.set_ylim(-50, 2)

    ax1.scatter(
        t, mwa_pass_fit, color="#4f8a8b", alpha=0.9, marker=".", label="RFE cali"
    )
    ax1.set_ylim(-50, 2)

    leg = ax1.legend(frameon=True)
    leg.get_frame().set_facecolor("grey")
    leg.get_frame().set_alpha(0.2)
    for l in leg.legendHandles:
        l.set_alpha(1)

    plt.title(f"Goodness of Fit: {pval} ")
    plt.tight_layout()
    plt.savefig(f"{out_dir}/{point}/{timestamp}_{sat}.png")
    plt.close()


def power_ephem(
    ref, tile, ali_file, chrono_file, sat_id, sat_chan, point, pow_thresh, timestamp
):

    """Create power, alt, az arrays at constant cadence"""

    # Read .npz aligned file
    ref_p, tile_p, times = read_aligned(ali_file=ali_file)

    # Scale noise floor to zero and determine noise threshold
    ref_p, ref_noise = noise_floor(sat_thresh, noi_thresh, ref_p)
    tile_p, tile_noise = noise_floor(sat_thresh, noi_thresh, tile_p)

    with open(chrono_file) as chrono:
        chrono_ephem = json.load(chrono)

        norad_list = [chrono_ephem[s]["sat_id"][0] for s in range(len(chrono_ephem))]

        norad_index = norad_list.index(sat_id)

        norad_ephem = chrono_ephem[norad_index]

        rise_ephem = norad_ephem["time_array"][0]
        set_ephem = norad_ephem["time_array"][-1]

        intvl = time_filter(rise_ephem, set_ephem, np.asarray(times))

        if intvl != None:

            w_start, w_stop = intvl

            # Slice [crop] the ref/tile/times arrays to the times of sat pass and extract sat_chan
            ref_c = ref_p[w_start : w_stop + 1, sat_chan]
            tile_c = tile_p[w_start : w_stop + 1, sat_chan]
            times_c = times[w_start : w_stop + 1]

            alt = np.asarray(norad_ephem["sat_alt"])
            az = np.asarray(norad_ephem["sat_az"])

            if (np.nanmax(ref_c - ref_noise) >= pow_thresh) and (
                np.nanmax(tile_c - tile_noise) >= pow_thresh
            ):

                # Apply noise criteria. In the window, where are ref_power and tile power
                # above their respective thresholds?
                if np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0].size != 0:
                    good_ref = ref_c[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]
                    good_tile = tile_c[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]
                    good_alt = alt[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]
                    good_az = az[
                        np.where((ref_c >= ref_noise) & (tile_c >= tile_noise))[0]
                    ]

                    if plots == "True":
                        plt_channel(
                            f"{out_dir}/pass_plots/{tile}_{ref}/{point}",
                            times_c,
                            ref_c,
                            tile_c,
                            ref_noise,
                            tile_noise,
                            sat_chan,
                            sat_id,
                            point,
                            timestamp,
                            point,
                        )

                    return [good_ref, good_tile, good_alt, good_az, times_c]

                else:
                    return 0

            else:
                return 0

        else:
            return 0


def project_tile_healpix(tile_pair):

    ref, tile = tile_pair

    pointings = ["0", "2", "4", "41"]

    if plots == "True":
        Path(f"{out_dir}/pass_plots/{tile}_{ref}/0").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/pass_plots/{tile}_{ref}/2").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/pass_plots/{tile}_{ref}/4").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/pass_plots/{tile}_{ref}/41").mkdir(parents=True, exist_ok=True)

        Path(f"{out_dir}/fit_plots/{tile}_{ref}/0").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/fit_plots/{tile}_{ref}/2").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/fit_plots/{tile}_{ref}/4").mkdir(parents=True, exist_ok=True)
        Path(f"{out_dir}/fit_plots/{tile}_{ref}/41").mkdir(parents=True, exist_ok=True)

    # Initialize an empty dictionary for tile data
    # The map is list of length 12288 of empty lists to append pixel values to
    # keep track of which satellites contributed which data
    tile_data = {
        "mwa_maps": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "ref_maps": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "tile_maps": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "sat_map": {
            p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings
        },
        "times": {p: [[] for pixel in range(hp.nside2npix(nside))] for p in pointings},
    }

    # Load reference FEE model
    # Rotate the fee models by -pi/2 to move model from spherical (E=0) to Alt/Az (N=0)
    ref_fee_model = np.load(ref_model, allow_pickle=True)

    #    # Tile S33YY has it's 9th dipole flagged
    #    if tile is 'S33YY':
    #        fee_m = np.load(fee_map_flagged, allow_pickle=True)
    #    else:
    #        fee_m = np.load(fee_map, allow_pickle=True)
    fee_m = np.load(fee_map, allow_pickle=True)

    if "XX" in tile:
        ref_fee = ref_fee_model["XX"]
        rotated_fee = rotate(nside, angle=-(1 * np.pi) / 2.0, healpix_array=ref_fee)
    else:
        ref_fee = ref_fee_model["YY"]
        rotated_fee = rotate(nside, angle=-(1 * np.pi) / 2.0, healpix_array=ref_fee)

    for day in range(len(dates)):

        for window in range(len(date_time[day])):
            timestamp = date_time[day][window]

            # Check if at timestamp, reciever was pointed to 0,2,4 gridpointing
            if (
                (timestamp in point_0)
                or (timestamp in point_2)
                or (timestamp in point_4)
                or (timestamp in point_41)
            ):

                # pointing at timestamp
                point = check_pointing(timestamp, point_0, point_2, point_4, point_41)

                if "XX" in tile:
                    mwa_fee = fee_m[str(point)][0]
                else:
                    mwa_fee = fee_m[str(point)][1]

                ali_file = Path(
                    f"{align_dir}/{dates[day]}/{timestamp}/{ref}_{tile}_{timestamp}_aligned.npz"
                )

                # check if file exists
                if ali_file.is_file():

                    # Chrono and map Ephemeris file
                    chrono_file = Path(f"{chrono_dir}/{timestamp}.json")
                    channel_map = f"{map_dir}/{timestamp}.json"

                    with open(chrono_file) as chrono:
                        chrono_ephem = json.load(chrono)

                        if chrono_ephem != []:

                            norad_list = [
                                chrono_ephem[s]["sat_id"][0]
                                for s in range(len(chrono_ephem))
                            ]

                            if norad_list != []:

                                with open(channel_map) as ch_map:
                                    chan_map = json.load(ch_map)

                                    chan_sat_ids = [
                                        int(i) for i in list(chan_map.keys())
                                    ]

                                    for sat in chan_sat_ids:

                                        chan = chan_map[f"{sat}"]

                                        sat_data = power_ephem(
                                            ref,
                                            tile,
                                            ali_file,
                                            chrono_file,
                                            sat,
                                            chan,
                                            point,
                                            pow_thresh,
                                            timestamp,
                                        )

                                        if sat_data != 0:

                                            (
                                                ref_power,
                                                tile_power,
                                                alt,
                                                az,
                                                times,
                                            ) = sat_data

                                            # Altitude is in deg while az is in radians
                                            # convert alt to radians
                                            # za - zenith angle
                                            alt = np.radians(alt)
                                            za = np.pi / 2 - alt
                                            az = np.asarray(az)

                                            # Now convert to healpix coordinates
                                            # healpix_index = hp.ang2pix(nside,θ, ɸ)
                                            healpix_index = hp.ang2pix(nside, za, az)

                                            # multiple data points fall within a single healpix pixel
                                            # find the unique pixels
                                            u = np.unique(healpix_index)

                                            ref_pass = np.array(
                                                [
                                                    np.nanmean(
                                                        ref_power[
                                                            np.where(
                                                                healpix_index == i
                                                            )[0]
                                                        ]
                                                    )
                                                    for i in u
                                                ]
                                            )
                                            tile_pass = np.array(
                                                [
                                                    np.nanmean(
                                                        tile_power[
                                                            np.where(
                                                                healpix_index == i
                                                            )[0]
                                                        ]
                                                    )
                                                    for i in u
                                                ]
                                            )
                                            times_pass = np.array(
                                                [
                                                    np.mean(
                                                        times[
                                                            np.where(healpix_index == i)
                                                        ][0]
                                                    )
                                                    for i in u
                                                ]
                                            )
                                            ref_fee_pass = np.array(
                                                [rotated_fee[i] for i in u]
                                            )
                                            mwa_fee_pass = np.array(
                                                [mwa_fee[i] for i in u]
                                            )

                                            # implement RFE gain corrections here
                                            rfe_polyfit = np.load(rfe_gain)
                                            gain_cal = np.poly1d(rfe_polyfit)
                                            # rfe_thresh = gain_cal.roots[0]
                                            rfe_thresh = max(gain_cal.roots)
                                            tile_pass_rfe = [
                                                i + gain_cal(i)
                                                if i >= rfe_thresh
                                                else i
                                                for i in tile_pass
                                            ]

                                            # magic here
                                            # the beam shape finally emerges
                                            # mwa_pass = np.array(tile_pass) - np.array(ref_pass) + np.array(ref_fee_pass)
                                            mwa_pass = (
                                                np.array(tile_pass_rfe)
                                                - np.array(ref_pass)
                                                + np.array(ref_fee_pass)
                                            )

                                            # fit the power level of the pass to the mwa_fee model using a single gain value
                                            offset = fit_gain(
                                                map_data=mwa_pass, fee=mwa_fee_pass
                                            )

                                            mwa_pass_fit = mwa_pass - offset[0]

                                            if mwa_pass_fit.size != 0:

                                                # determine how well the data fits the model with chi-square
                                                pval = fit_test(
                                                    map_data=mwa_pass_fit,
                                                    fee=mwa_fee_pass,
                                                )

                                                if plots == "True":
                                                    mwa_pass_raw = (
                                                        np.array(tile_pass)
                                                        - np.array(ref_pass)
                                                        + np.array(ref_fee_pass)
                                                    )
                                                    offset = fit_gain(
                                                        map_data=mwa_pass_raw,
                                                        fee=mwa_fee_pass,
                                                    )
                                                    mwa_pass_fit_raw = (
                                                        mwa_pass_raw - offset[0]
                                                    )

                                                    plt_fee_fit(
                                                        times_pass,
                                                        mwa_fee_pass,
                                                        mwa_pass_fit,
                                                        mwa_pass_fit_raw,
                                                        f"{out_dir}/fit_plots/{tile}_{ref}/",
                                                        point,
                                                        timestamp,
                                                        sat,
                                                    )

                                                # a goodness of fit threshold
                                                if pval >= 0.8:

                                                    # loop though all healpix pixels for the pass
                                                    for i in range(len(u)):

                                                        tile_data["mwa_maps"][
                                                            f"{point}"
                                                        ][u[i]].append(mwa_pass_fit[i])
                                                        tile_data["ref_maps"][
                                                            f"{point}"
                                                        ][u[i]].append(ref_pass[i])
                                                        tile_data["tile_maps"][
                                                            f"{point}"
                                                        ][u[i]].append(tile_pass[i])
                                                        tile_data["times"][f"{point}"][
                                                            u[i]
                                                        ].append(times_pass[i])
                                                        tile_data["sat_map"][
                                                            f"{point}"
                                                        ][u[i]].append(sat)

                else:
                    print(f"Missing {ref}_{tile}_{timestamp}_aligned.npz")
                    continue

    # Sort data by satellites

    # list of all possible satellites
    sat_ids = list(norad_ids.values())

    # create a dictionary, with the keys being sat_ids and the values being healpix maps of data from those sats
    # Fist level keys are pointings with dictionaty values
    # Second level keys are satellite ids with healpix map lists
    mwa_sat_data = {}
    ref_sat_data = {}
    tile_sat_data = {}
    time_sat_data = {}

    # loop over pointings for all maps
    for p in pointings:

        mwa_map = np.asarray(tile_data["mwa_maps"][p])
        tile_map = np.asarray(tile_data["tile_maps"][p])
        ref_map = np.asarray(tile_data["ref_maps"][p])
        sat_map = np.asarray(tile_data["sat_map"][p])
        time_map = np.asarray(tile_data["times"][p])

        mwa_sat_data[p] = {}
        ref_sat_data[p] = {}
        tile_sat_data[p] = {}
        time_sat_data[p] = {}

        # loop over all sats
        for s in sat_ids:

            mwa_sat_data[p][s] = []
            ref_sat_data[p][s] = []
            tile_sat_data[p][s] = []
            time_sat_data[p][s] = []

            # loop over every healpix pixel
            for i in range(len(ref_map)):

                # Find subset of data for each sat
                sat_idx = np.where(np.asarray(sat_map[i]) == s)
                mwa_sat_data[p][s].append((np.asarray(mwa_map[i])[sat_idx]).tolist())
                ref_sat_data[p][s].append((np.asarray(ref_map[i])[sat_idx]).tolist())
                tile_sat_data[p][s].append((np.asarray(tile_map[i])[sat_idx]).tolist())
                time_sat_data[p][s].append((np.asarray(time_map[i])[sat_idx]).tolist())

    # Save map arrays to npz file
    tile_sat_data = {
        "mwa_map": mwa_sat_data,
        "ref_map": ref_sat_data,
        "tile_map": tile_sat_data,
        "time_map": time_sat_data,
    }
    np.savez_compressed(f"{out_dir}/{tile}_{ref}_sat_maps.npz", **tile_sat_data)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="""
        Project a sat pass onto a healpix map using ephemeris data
        """
    )

    parser.add_argument(
        "--align_dir",
        metavar="\b",
        default="../../outputs/align_data/",
        help="Dir where agligned data date is saved. Default:../../outputs/align_data",
    )

    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/tile_maps/tile_maps_raw/",
        help="Output directory. Default=./../../outputs/tile_maps/tile_maps_raw/",
    )

    parser.add_argument(
        "--obs_point",
        metavar="\b",
        default="../../outputs/beam_pointings/obs_pointings.json",
        help="Observation pointing lists. Default=../../outputs/beam_pointings/obs_pointings.json",
    )

    parser.add_argument(
        "--chrono_dir",
        metavar="\b",
        default="./../../outputs/sat_ephemeris/chrono_json",
        help="Output directory. Default=./../../outputs/sat_ephemeris/chrono_json/",
    )

    parser.add_argument(
        "--map_dir",
        metavar="\b",
        default="../../outputs/sat_channels/window_maps/",
        help="Satellite channel map. Default=../../outputs/sat_channels/window_maps/",
    )

    parser.add_argument(
        "--ref_model",
        metavar="\b",
        default="../../outputs/reproject_ref/ref_dipole_models.npz",
        help="Healpix reference FEE model file. default=../../outputs/reproject_ref/ref_dipole_models.npz",
    )

    parser.add_argument(
        "--rfe_gain",
        metavar="\b",
        default="../../outputs/tile_maps/rfe_gain/rfe_gain_fit.npy",
        help="RF Explorer gain fit. default=../../outputs/tile_maps/rfe_gain/rfe_gain_fit.npy",
    )

    parser.add_argument(
        "--fee_map",
        metavar="\b",
        default="../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz",
        help="Healpix FEE map of mwa tile. default=../../outputs/tile_maps/FEE_maps/mwa_fee_beam.npz",
    )

    parser.add_argument(
        "--fee_map_flagged",
        metavar="\b",
        default="../../outputs/tile_maps/FEE_maps/mwa_fee_beam_9_flagged.npz",
        help="Healpix FEE map of mwa tile. default=../../outputs/tile_maps/FEE_maps/mwa_fee_beam_9_flagged.npz",
    )

    parser.add_argument(
        "--fit_thresh",
        metavar="\b",
        default=0.9,
        help="Goodness of fit threshold. 0.9 seems to only reject obvious outliers",
    )
    parser.add_argument(
        "--noi_thresh",
        metavar="\b",
        type=int,
        default=3,
        help="Noise Threshold: Multiples of MAD. Default=3.",
    )
    parser.add_argument(
        "--sat_thresh",
        metavar="\b",
        type=int,
        default=1,
        help="σ threshold to detect sats Default=1.",
    )
    parser.add_argument(
        "--pow_thresh",
        metavar="\b",
        type=int,
        default=5,
        help="Power Threshold to detect sats. Default=10 dB.",
    )
    parser.add_argument(
        "--nside",
        metavar="\b",
        type=int,
        default=32,
        help="Healpix Nside. Default = 32",
    )
    parser.add_argument(
        "--plots",
        metavar="\b",
        default=False,
        help="If True, create a gazzillion plots for each sat pass. Default = False",
    )
    parser.add_argument(
        "--start_date",
        metavar="\b",
        help="Date from which to start aligning data. Ex: 2019-10-10",
    )
    parser.add_argument(
        "--stop_date",
        metavar="\b",
        help="Date until which to align data. Ex: 2019-10-11",
    )

    args = parser.parse_args()

    start_date = args.start_date
    stop_date = args.stop_date
    obs_point = args.obs_point
    noi_thresh = args.noi_thresh
    sat_thresh = args.sat_thresh
    pow_thresh = args.pow_thresh
    fit_thresh = args.fit_thresh
    nside = args.nside
    plots = args.plots
    ref_model = args.ref_model
    rfe_gain = args.rfe_gain
    fee_map = args.fee_map
    fee_map_flagged = args.fee_map_flagged

    align_dir = Path(args.align_dir)
    chrono_dir = Path(args.chrono_dir)
    out_dir = Path(args.out_dir)
    map_dir = Path(args.map_dir)

    # Tile names
    refs = tile_names()[:4]
    tiles = tile_names()[4:]

    # All relevant tile pairs
    tile_pairs = []
    for ref in refs:
        if "XX" in ref:
            for tile in [t for t in tiles if "XX" in t]:
                tile_pairs.append([ref, tile])
        else:
            for tile in [t for t in tiles if "YY" in t]:
                tile_pairs.append([ref, tile])

    # Read observation pointing list
    with open(obs_point) as point:
        obs_p = json.load(point)
        point_0 = obs_p["point_0"]
        point_2 = obs_p["point_2"]
        point_4 = obs_p["point_4"]
        point_41 = obs_p["point_41"]

    # dates: list of days
    # date_time = list of 30 min observation windows
    dates, date_time = time_tree(start_date, stop_date)

    # Save logs
    Path(out_dir).mkdir(parents=True, exist_ok=True)
    sys.stdout = open(f"{out_dir}/logs_{start_date}_{stop_date}.txt", "a")

    #    project_tile_healpix(tile_pairs[0])
    #    project_tile_healpix(['rf0YY', 'S33YY'])

    # Parallization magic happens here
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = executor.map(project_tile_healpix, tile_pairs)



import sys
import json
import argparse
import numpy as np
import healpy as hp
from pathlib import Path
import concurrent.futures
from itertools import repeat
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
from scipy.stats import median_absolute_deviation as mad
from channels_plt import plt_waterfall_pass, plt_channel_basic, sat_plot


def time_tree(start_date, stop_date):
    """Split the time interval into 30 min obs"""

    t_start = datetime.strptime(start_date, "%Y-%m-%d")
    t_stop = datetime.strptime(stop_date, "%Y-%m-%d")
    n_days = (t_stop - t_start).days

    dates = []
    date_time = []

    # Every Day
    for i in range(n_days + 1):
        day = t_start + timedelta(days=i)
        date = day.strftime("%Y-%m-%d")
        dates.append(date)
        d_t = []

        # Every 30 min in the day
        for j in range(48):
            t_delta = datetime.strptime(date, "%Y-%m-%d") + timedelta(minutes=30 * j)
            d_time = t_delta.strftime("%Y-%m-%d-%H:%M")
            d_t.append(d_time)

        date_time.append(d_t)

    return (dates, date_time)


def read_aligned(ali_file=None):
    """Read aligned data npz file"""
    paired_data = np.load(ali_file, allow_pickle=True)

    power = paired_data["ref_p_aligned"]
    times = paired_data["time_array"]

    return [power, times]


def noise_floor(sat_thresh, noi_thresh, data=None):
    """Computes """

    # compute the standard deviation of data, and use it to identify occupied channels
    σ = np.std(data)

    # Any channel with a max power >= σ has a satellite
    sat_cut = sat_thresh * σ
    chans_pow_max = np.amax(data, axis=0)

    # Exclude the channels with sats, to only have noise data
    noise_chans = np.where(chans_pow_max < sat_cut)[0]
    noise_data = data[:, noise_chans]

    # noise median, noise mad, noise threshold = μ + 3*σ
    μ_noise = np.median(noise_data)
    σ_noise = mad(noise_data, axis=None)
    # because we rescale power to have zero median
    noise_threshold = (μ_noise - μ_noise) + noi_thresh * σ_noise

    # scale the data so that it has zero median
    data = data - μ_noise

    return (data, noise_threshold)


def time_filter(s_rise, s_set, times):
    """Slice obs window to size of ephem of norad sat
    Args:
        s_rise          : ephem rise time
        s_set           : ephem set time
        times           : time array of obs

    Returns:
        intvl           : index of rf time array where sat is up
        """

    # I. sat rises before times, sets within times window
    if s_rise < times[0] and s_set > times[0] and s_set <= times[-1]:
        i_0 = np.where(times == times[0])[0][0]
        i_1 = np.where(times == s_set)[0][0]
        intvl = [i_0, i_1]

    # II. sat rises and sets within times
    elif s_rise >= times[0] and s_set <= times[-1]:
        i_0 = np.where(times == s_rise)[0][0]
        i_1 = np.where(times == s_set)[0][0]
        intvl = [i_0, i_1]

    # III. sat rises within times, and sets after
    elif s_rise >= times[0] and s_rise < times[-1] and s_set > times[-1]:
        i_0 = np.where(times == s_rise)[0][0]
        i_1 = np.where(times == times[-1])[0][0]
        intvl = [i_0, i_1]

    # IV. sat rises before times and sets after
    elif s_rise < times[0] and s_set > times[-1]:
        i_0 = np.where(times == times[0])[0][0]
        i_1 = np.where(times == times[-1])[0][0]
        intvl = [i_0, i_1]

    # V. sat completely out of times. Could be on either side
    else:
        intvl = None

    # intvl = interval
    return intvl


def good_chans(
    ref_file,
    chrono_file,
    sat_id,
    sat_thresh,
    noi_thresh,
    pow_thresh,
    occ_thresh,
    date,
    timestamp,
    plots,
):

    """Create power, alt, az arrays at constant cadence"""

    # Custom spectral colormap
    import sys

    sys.path.append("../decode_rf_data")
    from colormap import spectral

    cmap = spectral()

    power, times = read_aligned(ali_file=ref_file)

    # Scale noise floor to zero and determine noise threshold
    power, noise_threshold = noise_floor(sat_thresh, noi_thresh, power)

    with open(chrono_file) as chrono:
        chrono_ephem = json.load(chrono)

        norad_list = [chrono_ephem[s]["sat_id"][0] for s in range(len(chrono_ephem))]

        norad_index = norad_list.index(sat_id)

        norad_ephem = chrono_ephem[norad_index]

        rise_ephem = norad_ephem["time_array"][0]
        set_ephem = norad_ephem["time_array"][-1]

        intvl = time_filter(rise_ephem, set_ephem, np.asarray(times))

        if intvl != None:

            w_start, w_stop = intvl

            window_len = w_stop - w_start + 1

            # Slice [crop] the power/times arrays to the times of sat pass
            power_c = power[w_start : w_stop + 1, :]
            times_c = times[w_start : w_stop + 1]

            possible_chans = []
            occu_list = []

            # Loop over every channel
            for s_chan in range(len(power_c[0])):

                channel_power = power_c[:, s_chan]

                # Percentage of signal occupancy above noise threshold
                window_occupancy = (np.where(channel_power >= noise_threshold))[
                    0
                ].size / window_len
                min_s = np.amin(channel_power)

                # Power threshold below which satellites aren't counted
                # Only continue if there is signal for more than 80% of satellite pass
                if (
                    max(channel_power) >= pow_thresh
                    and occ_thresh <= window_occupancy < 1.00
                ):

                    if times[0] == times_c[0]:
                        if (
                            all(p < noise_threshold for p in channel_power[-11:-1])
                        ) is True:
                            occu_list.append(window_occupancy)
                            possible_chans.append(s_chan)
                            # print(f'{sat_id}: {s_chan} {window_occupancy}')

                            if plots == "True":
                                plt_channel_basic(
                                    f"{plt_dir}/{date}/{timestamp}",
                                    times_c,
                                    channel_power,
                                    s_chan,
                                    min_s,
                                    32,
                                    noise_threshold,
                                    pow_thresh,
                                    sat_id,
                                    timestamp,
                                )

                    elif times[-1] == times_c[-1]:
                        if (
                            all(p < noise_threshold for p in channel_power[:10])
                        ) is True:
                            occu_list.append(window_occupancy)
                            possible_chans.append(s_chan)
                            # print(f'{sat_id}: {s_chan} {window_occupancy}')

                            if plots == "True":
                                plt_channel_basic(
                                    f"{plt_dir}/{date}/{timestamp}",
                                    times_c,
                                    channel_power,
                                    s_chan,
                                    min_s,
                                    32,
                                    noise_threshold,
                                    pow_thresh,
                                    sat_id,
                                    timestamp,
                                )

                    else:
                        if (
                            all(p < noise_threshold for p in channel_power[:10])
                            and all(p < noise_threshold for p in channel_power[-11:-1])
                        ) is True:
                            occu_list.append(window_occupancy)
                            possible_chans.append(s_chan)
                            # print(f'{sat_id}: {s_chan} {window_occupancy}')

                            if plots == "True":
                                plt_channel_basic(
                                    f"{plt_dir}/{date}/{timestamp}",
                                    times_c,
                                    channel_power,
                                    s_chan,
                                    min_s,
                                    32,
                                    noise_threshold,
                                    pow_thresh,
                                    sat_id,
                                    timestamp,
                                )

            # If channels are identified in the 30 min obs
            n_chans = len(possible_chans)

            if n_chans > 0:

                # The most lightly channel is one with the highest occupation
                good_chan = possible_chans[occu_list.index(max(occu_list))]

                if plots == "True":
                    plt_waterfall_pass(
                        power,
                        sat_id,
                        w_start,
                        w_stop,
                        timestamp,
                        cmap,
                        chs=possible_chans,
                        good_ch=good_chan,
                        out_dir=f"{plt_dir}/{date}/{timestamp}",
                    )

                return good_chan

            else:
                if plots == "True":
                    plt_waterfall_pass(
                        power,
                        sat_id,
                        w_start,
                        w_stop,
                        timestamp,
                        cmap,
                        chs=None,
                        good_ch=None,
                        out_dir=f"{plt_dir}/{date}/{timestamp}",
                    )
                return 0

        else:
            return 0


def window_chan_map(
    ali_dir,
    chrono_dir,
    sat_thresh,
    noi_thresh,
    pow_thresh,
    occ_thresh,
    out_dir,
    plots,
    plt_dir,
    obs_stamp,
):

    date, timestamp = obs_stamp

    if plots == "True":
        Path(f"{plt_dir}/{date}/{timestamp}").mkdir(parents=True, exist_ok=True)

    channel_map = {}

    ref_file = f"{ali_dir}/{date}/{timestamp}/rf0XX_S07XX_{timestamp}_aligned.npz"

    if Path(ref_file).is_file():
        ref_file = f"{ali_dir}/{date}/{timestamp}/rf0XX_S07XX_{timestamp}_aligned.npz"
    elif Path(
        ref_file=f"{ali_dir}/{date}/{timestamp}/rf0XX_S36XX_{timestamp}_aligned.npz"
    ).is_file():
        ref_file = f"{ali_dir}/{date}/{timestamp}/rf0XX_S36XX_{timestamp}_aligned.npz"
    else:
        ref_file = f"{ali_dir}/{date}/{timestamp}/rf0XX_S06XX_{timestamp}_aligned.npz"

    chrono_file = f"{chrono_dir}/{timestamp}.json"

    try:
        Path(ref_file).is_file()

        with open(chrono_file) as chrono:
            chrono_ephem = json.load(chrono)

            if chrono_ephem != []:

                norad_list = [
                    chrono_ephem[s]["sat_id"][0] for s in range(len(chrono_ephem))
                ]

                if norad_list != []:

                    for sat in norad_list:

                        sat_data = good_chans(
                            ref_file,
                            chrono_file,
                            sat,
                            sat_thresh,
                            noi_thresh,
                            pow_thresh,
                            occ_thresh,
                            date,
                            timestamp,
                            plots,
                        )

                        if sat_data != 0:
                            sat_chan = sat_data

                            channel_map[f"{sat}"] = sat_chan

    except Exception as e:
        print(e)
        # Exception message is forwarded from ../decode_rf_data/rf_data.py

    # Save channel map
    Path(f"{out_dir}/window_maps").mkdir(parents=True, exist_ok=True)
    with open(f"{out_dir}/window_maps/{timestamp}.json", "w") as f:
        json.dump(channel_map, f, indent=4)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="""
        Project a sat pass onto a healpix map using ephemeris data
        """
    )

    parser.add_argument(
        "--start_date",
        metavar="\b",
        help="Date from which to start aligning data. Ex: 2019-10-10",
    )
    parser.add_argument(
        "--stop_date",
        metavar="\b",
        help="Date until which to align data. Ex: 2019-10-11",
    )
    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/sat_channels/",
        help="Output directory. Default=./../../outputs/sat_channels/",
    )
    parser.add_argument(
        "--plt_dir",
        metavar="\b",
        default="./../../outputs/sat_channels/window_plots",
        help="Output directory. Default=./../../outputs/sat_channels/window_plots/",
    )
    parser.add_argument(
        "--ali_dir",
        metavar="\b",
        default="./../../outputs/align_data/",
        help="Output directory. Default=./../../outputs/align_data/",
    )
    parser.add_argument(
        "--chrono_dir",
        metavar="\b",
        default="./../../outputs/sat_ephemeris/chrono_json",
        help="Output directory. Default=./../../outputs/sat_ephemeris/chrono_json/",
    )
    parser.add_argument(
        "--noi_thresh",
        metavar="\b",
        type=int,
        default=3,
        help="Noise Threshold: Multiples of MAD. Default=3.",
    )
    parser.add_argument(
        "--sat_thresh",
        metavar="\b",
        type=int,
        default=1,
        help="1 σ threshold to detect sats Default=1.",
    )
    parser.add_argument(
        "--pow_thresh",
        metavar="\b",
        type=int,
        default=15,
        help="Power Threshold to detect sats. Default=15 dB.",
    )
    parser.add_argument(
        "--occ_thresh",
        metavar="\b",
        type=int,
        default=0.80,
        help="Occupation Threshold of sat in window. Default=0.80",
    )
    parser.add_argument(
        "--plots",
        metavar="\b",
        default=False,
        help="If True, create a gazzillion plots for each sat pass. Default = False",
    )

    args = parser.parse_args()

    chrono_dir = args.chrono_dir
    start_date = args.start_date
    stop_date = args.stop_date
    out_dir = args.out_dir
    plt_dir = args.plt_dir
    ali_dir = args.ali_dir
    noi_thresh = args.noi_thresh
    sat_thresh = args.sat_thresh
    pow_thresh = args.pow_thresh
    occ_thresh = args.occ_thresh
    plots = args.plots

    ref_names = ["rf0XX", "rf0YY", "rf1XX", "rf1YY"]

    # Save logs
    Path(f"{out_dir}/window_maps").mkdir(parents=True, exist_ok=True)
    sys.stdout = open(f"{out_dir}/logs_{start_date}_{stop_date}.txt", "a")

    # Help traverse all 30 min obs b/w start & stop
    dates, date_time = time_tree(start_date, stop_date)
    obs_list = [
        [dates[d], date_time[d][dt]]
        for d in range(len(dates))
        for dt in range(len(date_time[d]))
    ]

    # Parallization magic happens here
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = executor.map(
            window_chan_map,
            repeat(ali_dir),
            repeat(chrono_dir),
            repeat(sat_thresh),
            repeat(noi_thresh),
            repeat(pow_thresh),
            repeat(occ_thresh),
            repeat(out_dir),
            repeat(plots),
            repeat(plt_dir),
            obs_list,
        )



from astropy.time import Time


def convert_gps(date):
    return int(Time(date, format="iso").gps)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="""
            Convert date-time [iso] to gps time
            """
    )
    parser.add_argument(
        "--date",
        metavar="\b",
        help="Date-time to be converted to gps-seconds. Ex:2019-01-01 00:00:00",
    )

    args = parser.parse_args()
    date = args.date

    gps = convert_gps(date)
    print(f"gps: {gps}")



import os
import time
import argparse
from pathlib import Path
from sat_ids import norad_ids
import spacetrack.operators as op
from spacetrack import SpaceTrackClient

# space-tracks.org login and password, saved as environment variables

if os.environ.get("ST_USER") != None:
    st_user = os.environ.get("ST_USER")
    st_pass = os.environ.get("ST_PASS")

    st = SpaceTrackClient(identity=st_user, password=st_pass)

    parser = argparse.ArgumentParser(
        description="""Downloads TLE files for the Beam Experiment from Space-tracks.org."""
    )
    parser.add_argument(
        "--start_date", metavar="\b", help="Start date in format YYYY-MM-DD"
    )
    parser.add_argument(
        "--stop_date", metavar="\b", help="Stop date in format YYYY-MM-DD"
    )
    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/sat_ephemeris/TLE/",
        help="Path to output directory. Default=./../../outputs/sat_ephemeris/TLE/",
    )

    args = parser.parse_args()

    start_date = args.start_date
    stop_date = args.stop_date
    out_dir = args.out_dir

    # make a TLE directory
    Path(out_dir).mkdir(parents=True, exist_ok=True)
    os.makedirs(os.path.dirname(out_dir), exist_ok=True)

    print("Starting TLE download")

    for sat_name, sat_id in norad_ids.items():
        # Sleep to limit downloads to 200 TLEs per hour
        time.sleep(20)
        data = st.tle(
            iter_lines=True,
            norad_cat_id=sat_id,
            orderby="epoch asc",
            epoch=f"{start_date}--{stop_date}",
            format="tle",
        )
        print(
            f"downloading tle for {sat_name} satellite [{sat_id}] from space-tracks.org"
        )
        with open(f"{out_dir}/{sat_id}.txt", "w") as fp:
            for line in data:
                fp.write(line + "\n")


else:
    print("ERROR!")
    print("Space-Track.org credentials do not exist!")
    print("Edit this file and enter your account details.")



import json
import numpy as np
import seaborn as sns
from pathlib import Path
import matplotlib.pyplot as plt

import sys

sys.path.append("../sat_channels")
from window_chan_map import time_tree


def point_int(
    tile, start_date, stop_date, point_0, point_2, point_4, point_41, data_dir
):
    """Calculate integration at each pointing for every tile"""

    # increment this by 0.5, for every obs which is in point_list.
    # 0.5 = 30 min
    p_0 = 0
    p_2 = 0
    p_4 = 0
    p_41 = 0

    # dates: list of days
    # date_time = list of 30 min observation windows
    dates, date_time = time_tree(start_date, stop_date)
    for day in range(len(dates)):
        for window in range(len(date_time[day])):
            f_name = f"{tile}_{date_time[day][window]}.txt"
            f_path = Path(f"{data_dir}/{tile}/{dates[day]}/{f_name}")
            if f_path.is_file():
                if date_time[day][window] in point_0:
                    p_0 += 0.5
                elif date_time[day][window] in point_2:
                    p_2 += 0.5
                elif date_time[day][window] in point_4:
                    p_4 += 0.5
                elif date_time[day][window] in point_41:
                    p_41 += 0.5
                else:
                    pass
    return [p_0, p_2, p_4, p_41]


def plt_pointing_hist(a, b, c, fig, int_dict=None, tile=None, x_ticks=None):

    ax = fig.add_subplot(a, b, c)

    y_max = max(i for v in int_dict.values() for i in v)
    int_list = int_dict[tile]
    x = range(len(int_list))
    leg = int_list

    pal = sns.cubehelix_palette(
        len(int_list), start=0.7, rot=-0.7, dark=0.4, reverse=True
    )
    barplot = plt.bar(x, int_list, color=sns.color_palette(pal), edgecolor="black")

    def autolabel(rects):
        for idx, rect in enumerate(barplot):
            height = rect.get_height()
            ax.text(
                rect.get_x() + rect.get_width() / 2.0,
                height,
                leg[idx],
                ha="center",
                va="bottom",
                rotation=0,
            )

    autolabel(barplot)

    # these are matplotlib.patch.Patch properties
    props = dict(boxstyle="round", facecolor="wheat", alpha=0.5)

    # place a text box in upper left in axes coords
    ax.text(
        0.65,
        0.95,
        f"{tile}",
        transform=ax.transAxes,
        fontsize=10,
        verticalalignment="top",
        bbox=props,
    )

    plt.xlim(-0.5, len(int_list) - 0.5)
    ax.set_yticklabels([])

    ax.set_xticklabels([])
    ax.set_ylim(0, 1.1 * y_max)

    return ax


if __name__ == "__main__":

    import sys
    import argparse

    sys.path.append("../decode_rf_data")
    from rf_data import tile_names

    parser = argparse.ArgumentParser(
        description="""
        Determine total integration time, per pointing, for all tiles
        """
    )

    parser.add_argument("--data_dir", metavar="\b", help="Data directory")
    parser.add_argument(
        "--out_dir",
        metavar="\b",
        default="./../../outputs/beam_pointings/",
        help="Output directory. Default=./../../outputs/beam_pointings/",
    )

    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    out_dir = Path(args.out_dir)

    # Tile names
    tiles = tile_names()

    # lists of obs at each pointings
    with open(f"{out_dir}/obs_pointings.json") as table:
        data = json.load(table)
        start_date = data["start_date"]
        stop_date = data["stop_date"]
        point_0 = data["point_0"]
        point_2 = data["point_2"]
        point_4 = data["point_4"]
        point_41 = data["point_41"]

    tile_integration = {}
    for tile in tiles:
        p_integration = point_int(
            tile, start_date, stop_date, point_0, point_2, point_4, point_41, data_dir
        )
        tile_integration[f"{tile}"] = p_integration

    plt.style.use("seaborn")
    fig = plt.figure(figsize=(14, 10))

    ax01 = plt_pointing_hist(4, 8, 1, fig, int_dict=tile_integration, tile=tiles[0])
    ax02 = plt_pointing_hist(4, 8, 2, fig, int_dict=tile_integration, tile=tiles[4])
    ax03 = plt_pointing_hist(4, 8, 3, fig, int_dict=tile_integration, tile=tiles[5])
    ax04 = plt_pointing_hist(4, 8, 4, fig, int_dict=tile_integration, tile=tiles[6])
    ax05 = plt_pointing_hist(4, 8, 5, fig, int_dict=tile_integration, tile=tiles[7])
    ax06 = plt_pointing_hist(4, 8, 6, fig, int_dict=tile_integration, tile=tiles[8])
    ax07 = plt_pointing_hist(4, 8, 7, fig, int_dict=tile_integration, tile=tiles[9])
    ax08 = plt_pointing_hist(4, 8, 8, fig, int_dict=tile_integration, tile=tiles[10])
    ax09 = plt_pointing_hist(4, 8, 9, fig, int_dict=tile_integration, tile=tiles[1])
    ax10 = plt_pointing_hist(4, 8, 10, fig, int_dict=tile_integration, tile=tiles[11])
    ax11 = plt_pointing_hist(4, 8, 11, fig, int_dict=tile_integration, tile=tiles[12])
    ax12 = plt_pointing_hist(4, 8, 12, fig, int_dict=tile_integration, tile=tiles[13])
    ax12 = plt_pointing_hist(4, 8, 13, fig, int_dict=tile_integration, tile=tiles[14])
    ax13 = plt_pointing_hist(4, 8, 14, fig, int_dict=tile_integration, tile=tiles[15])
    ax14 = plt_pointing_hist(4, 8, 15, fig, int_dict=tile_integration, tile=tiles[16])
    ax15 = plt_pointing_hist(4, 8, 16, fig, int_dict=tile_integration, tile=tiles[17])
    ax16 = plt_pointing_hist(4, 8, 17, fig, int_dict=tile_integration, tile=tiles[2])
    ax17 = plt_pointing_hist(4, 8, 18, fig, int_dict=tile_integration, tile=tiles[18])
    ax18 = plt_pointing_hist(4, 8, 19, fig, int_dict=tile_integration, tile=tiles[19])
    ax19 = plt_pointing_hist(4, 8, 20, fig, int_dict=tile_integration, tile=tiles[20])
    ax20 = plt_pointing_hist(4, 8, 21, fig, int_dict=tile_integration, tile=tiles[21])
    ax21 = plt_pointing_hist(4, 8, 22, fig, int_dict=tile_integration, tile=tiles[22])
    ax22 = plt_pointing_hist(4, 8, 23, fig, int_dict=tile_integration, tile=tiles[23])
    ax23 = plt_pointing_hist(4, 8, 24, fig, int_dict=tile_integration, tile=tiles[24])
    ax24 = plt_pointing_hist(4, 8, 25, fig, int_dict=tile_integration, tile=tiles[3])
    ax25 = plt_pointing_hist(4, 8, 26, fig, int_dict=tile_integration, tile=tiles[25])
    ax26 = plt_pointing_hist(4, 8, 27, fig, int_dict=tile_integration, tile=tiles[26])
    ax27 = plt_pointing_hist(4, 8, 28, fig, int_dict=tile_integration, tile=tiles[27])
    ax28 = plt_pointing_hist(4, 8, 29, fig, int_dict=tile_integration, tile=tiles[28])
    ax29 = plt_pointing_hist(4, 8, 30, fig, int_dict=tile_integration, tile=tiles[29])
    ax30 = plt_pointing_hist(4, 8, 31, fig, int_dict=tile_integration, tile=tiles[30])
    ax31 = plt_pointing_hist(4, 8, 32, fig, int_dict=tile_integration, tile=tiles[31])

    #    plt.title('Pointing Integration per Tile')
    #    plt.ylabel('Hours')
    #    plt.xlabel('Pointing [0, 2, 4]')
    #    fig.text(0.5, 0.04, 'MWA Receiver Pointing [0, 2, 4]', ha='center')
    #    fig.text(0.04, 0.5, 'Hours', va='center', rotation='vertical')
    plt.tight_layout()
    fig.savefig(f"{out_dir}/tiles_pointing_integration.png")



"""
Align Data
==========
Tools to time align pairs of rf data files,
enabling comparisons between corresponding 
data points.

"""

import re
import math
import time
import numpy as np
from pathlib import Path
from scipy import interpolate
from scipy.signal import savgol_filter
from embers.condition_data.rf_data import read_data


def savgol_interp(
    ref,
    tile,
    savgol_window_1=None,
    savgol_window_2=None,
    polyorder=None,
    interp_type=None,
    interp_freq=None,
):
    """Interpolate the power array followed by savgol smoothing.

    Interpolate to a given frequency which 
    makes the dimensions of the power arrays
    from the reference and tile antennas equal,
    enabling comparisons between corresponding
    data points. Two level of savgol filter applied, 
    first to capture deep nulls + small structure, 
    and second level to smooth over noise.

    
    Parameters
    ----------
    :param str ref: path to reference data file
    :param str tile: path to tile data file
    :param int savgol_window_1:  window size of savgol filer, must be odd
    :param int savgol_window_2:  window size of savgol filer, must be odd
    :param int polyorder: polynomial order to fit to savgol_window
    :param str interp_type: type of interpolation. Ex: 'cubic', 'linear'
    :param int interp_freq: freqency to which power array is interpolated

    Returns
    -------
    :returns:
        - ref_ali - aligned reference power array
        - tile_ali - aligned tile power array
        - time_array - time array corresponding to power arrays
        - ref_power - raw reference power array
        - tile_power - raw tile power array
        - ref_time - raw reference time array
        - tile_time - raw tile time array

    :rtype: (float, numpy.array(float))

    """

    # Read time and power arrays from data files
    ref_power, ref_time = read_data(ref)
    tile_power, tile_time = read_data(tile)

    # Round up/down to nearest integer of time
    start_time = math.ceil(max(ref_time[0], tile_time[0]))
    stop_time = math.floor(min(ref_time[-1], tile_time[-1]))

    # Array of times at which to evaluate the interpolated data
    time_array = np.arange(start_time, stop_time, (1 / interp_freq))

    # Mathematical interpolation functions
    f = interpolate.interp1d(ref_time, ref_power, axis=0, kind=interp_type)
    g = interpolate.interp1d(tile_time, tile_power, axis=0, kind=interp_type)

    # New power array, evaluated at the desired frequency
    ref_ali = f(time_array)
    tile_ali = g(time_array)

    # Savgol level 1. Capture nulls / small scale structure
    ref_ali = savgol_filter(ref_ali, savgol_window_1, polyorder, axis=0)
    tile_ali = savgol_filter(tile_ali, savgol_window_1, polyorder, axis=0)

    # Savgol level 2. Smooth noise
    ref_ali = savgol_filter(ref_ali, savgol_window_2, polyorder, axis=0)
    tile_ali = savgol_filter(tile_ali, savgol_window_2, polyorder, axis=0)

    return (ref_ali, tile_ali, time_array, ref_power, tile_power, ref_time, tile_time)


def save_aligned(
    tile_pair,
    time_stamp,
    savgol_window_1,
    savgol_window_2,
    polyorder,
    interp_type,
    interp_freq,
    data_dir,
    out_dir,
):
    """Save an aligned set of rf data to an npz file. 

    A pair of rf data files are smoothed, interpolated and aligned
    with the :func:`~embers.condition_data.align_data.savgol_interp`.
    The output is written to a npz file and saved to an output 
    directory tree.

    Parameters
    ----------
    :param list[str] tile_pair: pair of ref and tile antenna names
    :param str time_stamp: time when rf observation began. In YYYY-MM-DD-HH-MM format
    :param int savgol_window_1:  window size of savgol filer, must be odd
    :param int savgol_window_2:  window size of savgol filer, must be odd
    :param int polyorder: polynomial order to fit to savgol_window
    :param str interp_type: type of interpolation. Ex: 'cubic', 'linear'
    :param int interp_freq: freqency to which power array is interpolated
    :param str data_dir: root of data dir where rf data is located
    :param str out_dir: relative path to output directory

    Returns
    -------
    :return: aligned rf data saved by `numpy.savez_compressed`
    
    Exeptions
    ---------
    :raises FileNotFoundError: an input file does not exist

    """

    date = re.search(r"\d{4}.\d{2}.\d{2}", time_stamp)[0]

    ref = tile_pair[0]
    tile = tile_pair[1]
    ref_file = f"{data_dir}/{ref}/{date}/{ref}_{time_stamp}.txt"
    tile_file = f"{data_dir}/{tile}/{date}/{tile}_{time_stamp}.txt"

    try:
        ref_ali, tile_ali, time_array, _, _, _, _ = savgol_interp(
            ref_file,
            tile_file,
            savgol_window_1=savgol_window_1,
            savgol_window_2=savgol_window_2,
            polyorder=polyorder,
            interp_type=interp_type,
            interp_freq=interp_freq,
        )

        # creates output directory if it doesn't exist
        save_dir = Path(f"{out_dir}/{date}/{time_stamp}")
        save_dir.mkdir(parents=True, exist_ok=True)

        # Convert the power array to float32
        # Convert list of times to float64 (double)
        # Save as compressed npz file. Seems to drastically reduce size
        np.savez_compressed(
            f"{save_dir}/{ref}_{tile}_{time_stamp}_aligned.npz",
            ref_ali=np.single(ref_ali),
            tile_ali=np.single(tile_ali),
            time_array=np.double(time_array),
        )

        return f"Saved aligned file to {save_dir}/{ref}_{tile}_{time_stamp}_aligned.npz"

    except FileNotFoundError as e:
        return e
    except Exception as e:
        return e



# Code developed by Jack Line and adapted here
# https://github.com/JLBLine/MWA_ORBCOMM

import numpy as np
import healpy as hp
from pathlib import Path
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from scipy import interpolate
from scipy.interpolate import RectSphereBivariateSpline, SmoothSphereBivariateSpline

import sys

sys.path.append("../decode_rf_data")
from colormap import spectral

cmap = spectral()


def create_model(len_empty_healpix, epsilon, nside, file_name=None):
    """Takes .ffe model, converts into healpix and smooths the response"""

    # Make an empty array for healpix projection
    beam_response = np.zeros(len_empty_healpix) * np.nan

    # Load ffe model data in, which is stored in real and imaginary components
    # of a phi/theta polaristaion representation of the beam
    # apparently this is normal to beam engineers
    data = np.loadtxt(file_name)
    theta = data[:, 0]
    phi = data[:, 1]
    re_theta = data[:, 2]
    im_theta = data[:, 3]
    re_phi = data[:, 4]
    im_phi = data[:, 5]

    # Convert to complex numbers
    X_theta = re_theta.astype(complex)
    X_theta.imag = im_theta

    X_phi = re_phi.astype(complex)
    X_phi.imag = im_phi

    # make a coord grid for the data in the .ffe model
    theta_range = np.linspace(epsilon, 90, 91)
    phi_range = np.linspace(epsilon, 359.0, 360)
    theta_mesh, phi_mesh = np.meshgrid(theta_range, phi_range)

    # Convert reference model into power from the complex values
    power = 10 * np.log10(abs(X_theta) ** 2 + abs(X_phi) ** 2)

    # Make an inf values super small
    power[np.where(power == -np.inf)] = -80

    # Had a problem with edge effects, leave off near horizon values
    # Basically remove the last 90 values of power list, where θ == 90
    power = power[:-91]

    # Get things into correct shape and do an interpolation
    power.shape = phi_mesh.shape

    # Bivariate spline approximation over a rectangular mesh on a sphere
    # s is a paramater I had to play with to get by eye nice results
    # s: positive smoothing factor
    lut = RectSphereBivariateSpline(
        theta_range * (np.pi / 180.0), phi_range * (np.pi / 180.0), power.T, s=0.1
    )

    # Get the theta and phi of all healpixels
    i_pix = np.arange(hp.nside2npix(nside))
    theta_rad, phi_rad = hp.pix2ang(nside, i_pix)

    # Evaluate the interpolated function at healpix gridpoints
    # Use spline to map beam into healpixels
    beam_response = lut.ev(theta_rad, phi_rad)

    # rescale beam response to have peak value of 0 dB
    beam_response = beam_response - np.nanmax(beam_response)

    return beam_response, theta_mesh, phi_mesh, power, theta


def plot_healpix(
    data_map=None,
    fig=None,
    sub=None,
    title=None,
    vmin=None,
    vmax=None,
    cmap=None,
    cbar=True,
):
    """Yeesh do some healpix magic to plot the thing"""

    # Disable cryptic healpy warnings. Can't figure out where they originate
    import warnings

    warnings.filterwarnings("ignore", category=RuntimeWarning)

    hp.orthview(
        map=data_map,
        coord="E",
        fig=fig,
        half_sky=True,
        rot=(0, 90, 180),
        xsize=600,
        title=title,
        sub=sub,
        min=vmin,
        max=vmax,
        cmap=cmap,
        notext=True,
        hold=True,
        cbar=cbar,
    )

    hp.graticule(dpar=10, coord="E", color="k", alpha=0.3, dmer=45)

    # Altitude grid
    hp.projtext(
        00.0 * (np.pi / 180.0), 225.0 * (np.pi / 180), "0", color="w", coord="E"
    )
    hp.projtext(
        30.0 * (np.pi / 180.0), 225.0 * (np.pi / 180), "30", color="w", coord="E"
    )
    hp.projtext(
        60.0 * (np.pi / 180.0), 225.0 * (np.pi / 180), "60", color="w", coord="E"
    )

    # NSEW
    hp.projtext(
        80.0 * (np.pi / 180.0),
        000.0 * (np.pi / 180.0),
        r"$N  $",
        coord="E",
        color="w",
        verticalalignment="top",
    )
    hp.projtext(
        80.0 * (np.pi / 180.0),
        090.0 * (np.pi / 180.0),
        r"$E  $",
        coord="E",
        color="w",
        horizontalalignment="right",
    )
    hp.projtext(
        80.0 * (np.pi / 180.0),
        180.0 * (np.pi / 180.0),
        r"$S  $",
        coord="E",
        color="w",
        verticalalignment="bottom",
    )
    hp.projtext(
        80.0 * (np.pi / 180.0),
        270.0 * (np.pi / 180.0),
        r"$W  $",
        coord="E",
        color="w",
        horizontalalignment="left",
    )


if __name__ == "__main__":

    # Healpix constants
    nside = 32
    len_empty_healpix = hp.nside2npix(nside)  # 12288

    # Had problems with having coords = 0 when interpolating, so make a small number
    # and call it epsilon for some reason

    epsilon = 1.0e-12

    # Reference FEE models in XX & YY pols
    refXX = "../../data/FEE_Reference_Models/MWA_reference_tile_FarField_XX.ffe"
    refYY = "../../data/FEE_Reference_Models/MWA_reference_tile_FarField_YY.ffe"

    # Output directory
    out_dir = "../../outputs/reproject_ref"
    Path(out_dir).mkdir(parents=True, exist_ok=True)

    healpix_XX, theta_mesh, phi_mesh, power_XX, theta = create_model(
        len_empty_healpix, epsilon, nside, file_name=refXX
    )
    healpix_YY, theta_mesh, phi_mesh, power_YY, theta = create_model(
        len_empty_healpix, epsilon, nside, file_name=refYY
    )

    # Plot the things to sanity check and save results
    fig = plt.figure(figsize=(10, 10))

    ax1 = fig.add_subplot(221, projection="polar")
    ax1.set_rgrids([10, 30, 50, 70, 90], angle=22)
    ax2 = fig.add_subplot(222, projection="polar")
    ax2.set_rgrids([10, 30, 50, 70, 90], angle=22)

    im1 = ax1.pcolormesh(
        phi_mesh * (np.pi / 180.0),
        theta_mesh,
        power_XX,
        label="XX",
        vmin=-40,
        vmax=-20,
        cmap=cmap,
    )

    ax1.set_title("Ref XX", y=1.1)
    ax1.grid(color="k", alpha=0.3)

    im2 = ax2.pcolormesh(
        phi_mesh * (np.pi / 180.0),
        theta_mesh,
        power_YY,
        label="YY",
        vmin=-40,
        vmax=-20,
        cmap=cmap,
    )

    ax2.set_title("Ref YY", y=1.1)
    ax2.grid(color="k", alpha=0.3)

    ax3 = fig.add_subplot(2, 2, 3)
    plot_healpix(
        data_map=healpix_XX, sub=(2, 2, 3), title=None, vmin=-20, vmax=0, cmap=cmap
    )
    ax3.axis("off")
    ax3.set_title("Healpix XX", x=0.4, y=-0.33)

    ax4 = fig.add_subplot(2, 2, 4)
    plot_healpix(
        data_map=healpix_YY, sub=(2, 2, 4), title=None, vmin=-20, vmax=0, cmap=cmap
    )
    ax4.axis("off")
    ax4.set_title("Healpix YY", x=0.6, y=-0.33)

    np.savez_compressed(
        f"{out_dir}/ref_dipole_models.npz", XX=healpix_XX, YY=healpix_YY
    )
    fig.savefig(f"{out_dir}/reproject_dipole_models.png", bbox_inches="tight")
